#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This is DVIasm, a DVI utility for editing DVI files directly.
#
# Copyright (C) 2007-2008 by Jin-Hwan Cho <chofchof@ktug.or.kr>
# Copyright (C) 2011-2017 by Khaled Hosny <khaledhosny@eglug.org>
# Copyright (C) 2019      by Arthur Reutenauer <arthur@reutenauer.eu>
# Copyright (C) 2019-2024 by Hironobu Yamashita <h.y.acetaminophen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys, os.path
from optparse import OptionParser

# Global variables
is_ptex = False
is_subfont = False
cur_font = None
cur_dsize = 0
cur_ssize = 0
subfont_idx = 0
subfont_list = ['cyberb', 'outbtm', 'outbtb', 'outgtm', 'outgtb']

# DVI opcodes
SET_CHAR_0 = 0; SET_CHAR_127 = 127;
SET1 = 128; SET2 = 129; SET3 = 130; SET4 = 131;
SET_RULE = 132;
PUT1 = 133; PUT2 = 134; PUT3 = 135; PUT4 = 136;
PUT_RULE = 137;
NOP = 138;
BOP = 139; EOP = 140;
PUSH = 141; POP = 142;
RIGHT1 = 143; RIGHT2 = 144; RIGHT3 = 145; RIGHT4 = 146;
W0 = 147; W1 = 148; W2 = 149; W3 = 150; W4 = 151;
X0 = 152; X1 = 153; X2 = 154; X3 = 155; X4 = 156;
DOWN1 = 157; DOWN2 = 158; DOWN3 = 159; DOWN4 = 160;
Y0 = 161; Y1 = 162; Y2 = 163; Y3 = 164; Y4 = 165;
Z0 = 166; Z1 = 167; Z2 = 168; Z3 = 169; Z4 = 170;
FNT_NUM_0 = 171; FNT_NUM_63 = 234;
FNT1 = 235; FNT2 = 236; FNT3 = 237; FNT4 = 238;
XXX1 = 239; XXX2 = 240; XXX3 = 241; XXX4 = 242;
FNT_DEF1 = 243; FNT_DEF2 = 244; FNT_DEF3 = 245; FNT_DEF4 = 246;
PRE = 247; POST = 248; POST_POST = 249;
# DVIV opcodes
DIR = 255;
# DVI-IVD opcodes
BEGIN_REFLECT = 250; END_REFLECT = 251;
# XDV opcodes
NATIVE_FONT_DEF = 252;
GLYPHS = 253;
TEXT_GLYPHS = 254;
# XDV flags
XDV_FLAG_VERTICAL = 0x0100;
XDV_FLAG_COLORED = 0x0200;
XDV_FLAG_EXTEND = 0x1000;
XDV_FLAG_SLANT = 0x2000;
XDV_FLAG_EMBOLDEN = 0x4000;
# DVI identifications
DVI_ID = 2; DVIV_ID = 3; XDVI_ID = 6; XDV_ID = 7;
DVI_IDS = (DVI_ID, DVIV_ID, XDVI_ID, XDV_ID)

def warning(msg):
  sys.stderr.write('warning: %s\n' % msg)

def ValidID(dvi_id):
    if dvi_id not in DVI_IDS:
        return False
    return True

def BadDVI(msg):
  raise AttributeError('Bad DVI file: %s!' % msg)

def GetByte(fp): # { returns the next byte, unsigned }
  try: return ord(fp.read(1))
  except: return -1

def SignedByte(fp): # { returns the next byte, signed }
  try: b = ord(fp.read(1))
  except: return -1
  if b < 128: return b
  else: return b - 256

def Get2Bytes(fp): # { returns the next two bytes, unsigned }
  try: a, b = fp.read(2)
  except: BadDVI('Failed to Get2Bytes()')
  return (a << 8) + b

def SignedPair(fp): # {returns the next two bytes, signed }
  try: a, b = fp.read(2)
  except: BadDVI('Failed to SignedPair()')
  if a < 128: return (a << 8) + b
  else: return ((a - 256) << 8) + b

def Get3Bytes(fp): # { returns the next three bytes, unsigned }
  try: a, b, c = fp.read(3)
  except: BadDVI('Failed to Get3Bytes()')
  return (((a << 8) + b) << 8) + c

def SignedTrio(fp): # { returns the next three bytes, signed }
  try: a, b, c = fp.read(3)
  except: BadDVI('Failed to SignedTrio()')
  if a < 128: return (((a << 8) + b) << 8) + c
  else: return ((((a - 256) << 8) + b) << 8) + c

def Get4Bytes(fp): # { returns the next four bytes, unsigned }
  try: a, b, c, d = fp.read(4)
  except: BadDVI('Failed to Get4Bytes()')
  return (((((a << 8) + b) << 8) + c) << 8) + d

def SignedQuad(fp): # { returns the next four bytes, signed }
  try: a, b, c, d = fp.read(4)
  except: BadDVI('Failed to get SignedQuad()')
  if a < 128: return (((((a << 8) + b) << 8) + c) << 8) + d
  else: return ((((((a - 256) << 8) + b) << 8) + c) << 8) + d

def PutByte(q):
  return bytes.fromhex('%02x' % (q & 0xff))

def Put2Bytes(q):
  return PutByte(q>>8) + PutByte(q)

def Put3Bytes(q):
  return PutByte(q>>16) + PutByte(q>>8) + PutByte(q)

def PutSignedQuad(q):
  if q < 0: q += 0x100000000
  return PutByte(q>>24) + PutByte(q>>16) + PutByte(q>>8) + PutByte(q)

def PutUnsigned(q):
  if q >= 0x1000000: return (3, PutSignedQuad(q))
  if q >= 0x10000:   return (2, Put3Bytes(q))
  if q >= 0x100:     return (1, Put2Bytes(q))
  return (0, PutByte(q))

def PutSigned(q):
  if q < -0x800000 or q >= 0x800000:  return (3, PutSignedQuad(q))
  if q >= 0x8000:                     return (2, Put3Bytes(q))
  if q < -0x8000:     q += 0x1000000; return (2, Put3Bytes(q))
  if q >= 0x80:                       return (1, Put2Bytes(q))
  if q < -0x80:       q += 0x10000;   return (1, Put2Bytes(q))
  return (0, PutByte(q))

def PutGlyphs(width, glyphs):
  s = []
  length = len(glyphs)
  s.append(PutSignedQuad(width))
  s.append(Put2Bytes(length))
  for glyph in glyphs:
    s.append(PutSignedQuad(glyph["x"]))
    s.append(PutSignedQuad(glyph["y"]))
  for glyph in glyphs:
    s.append(Put2Bytes(glyph["id"]))

  return b''.join(s)

def PutTextGlyphs(text, width, glyphs):
  s = []
  length = len(text)
  s.append(Put2Bytes(length))
  for ch in text:
    s.append(Put2Bytes(ch))
  s.append(PutGlyphs(width, glyphs))

  return b''.join(s)

def GetInt(s):
  try: return int(s)
  except: return -1

def GetStrASCII(s): # used in Parse()
  if len(s) > 1 and ((s[0] == "'" and s[-1] == "'") or (s[0] == '"' and s[-1] == '"')):
    return [ord(c) for c in s[1:-1].decode('unicode_escape')]
  else: return ''

def UCS2toJIS(c):
  try:
    s = c.encode('iso2022-jp')
  except UnicodeEncodeError:
    s = c.encode('raw_unicode_escape')
  if len(s) == 1: return ord(s)
  else:           return (s[3] << 8) + s[4]

def XUnicodeDecode(s): # dirty hack to handle >= 0x110000
  t = []
  i = 1
  while i < len(s)-1:
    c = -1
    if s[i] == '\\':
      if s[i+1] == 'x':
        try:
          c = int(s[i+2:i+4],16)
          i += 4
        except ValueError:
          warning('Invalid escape sequence \\x ignored!')
          i += 2
      elif s[i+1] == 'u':
        try:
          c = int(s[i+2:i+6],16)
          i += 6
        except ValueError:
          warning('Invalid escape sequence \\u ignored!')
          i += 2
      elif s[i+1] == 'U':
        try:
          c = int(s[i+2:i+10],16)
          i += 10
        except ValueError:
          warning('Invalid escape sequence \\U ignored!')
          i += 2
      else:
        if i+1 < len(s)-1:
          c = ord(s[i+1])
          i += 2
        else:
          warning('Invalid escape character \\ ignored!')
          i += 1
    else:
          c = ord(s[i])
          i += 1
    if is_ptex: c = UCS2toJIS(chr(c))
    if c >= 0: t.append(c)
  return t

def GetStrUTF8(s): # used in Parse()
  if len(s) > 1 and ((s[0] == "'" and s[-1] == "'") or (s[0] == '"' and s[-1] == '"')):
    try: # This should work for valid Unicode
      t = s[1:-1].encode('raw_unicode_escape').decode('unicode_escape')
      if is_ptex: return [UCS2toJIS(c) for c in t]
      else:       return [ord(c)       for c in t]
    except UnicodeDecodeError: # e.g. for upTeX >= 0x110000
                  return XUnicodeDecode(s)
  else:           return ''

def PutStrASCII(t): # used in Dump()
  s = ''
  for o in t:
    if o == 92:         s += '\\\\'
    elif 32 <= o < 127: s += chr(o)
    elif o < 256:       s += ('\\x%02x' % o)
    elif o < 65536:     s += ('\\u%04x' % o)
    else:               s += ('\\U%08x' % o)
  return "'%s'" % s

def PutStrLatin1(t): # used in Dump()
  s = ''
  for o in t:
    if o == 92:                           s += '\\\\'
    elif 32 <= o < 127 or 161 <= o < 256: s += chr(o)
    elif o < 256:                         s += ('\\x%02x' % o)
    elif o < 65536:                       s += ('\\u%04x' % o)
    else:                                 s += ('\\U%08x' % o)
  return "'%s'" % s

def DecodeISO2022JP(c):
  try:
    s = bytes.fromhex("1b 24 42 %02x %02x" % (c//256, c%256)).decode('iso2022-jp')
  except UnicodeDecodeError:
    s = chr(c)
  return s

def PutStrUTF8(t): # used in Dump()
  s = ''
  if is_subfont:
    for o in t:
      s += chr((subfont_idx << 8) + o)
  else: # not the case of subfont
    for o in t:
      if o == 92:         s += '\\\\'
      elif 32 <= o < 127: s += chr(o)
      elif o < 128:       s += ('\\x%02x' % o)
      elif is_ptex:
        s += DecodeISO2022JP(o)
      else:
        try:              s += chr(o)
        except ValueError: # upTeX may have >= 0x110000
          if o < 256:         s += '\\x%02x' % o
          elif o < 65536:     s += '\\u%04x' % o
          else:               s += '\\U%08x' % o
  return "'%s'" % s

def IsFontChanged(f, z):
  global cur_font, cur_ssize, subfont_idx, is_subfont
  for n in subfont_list:
    if n == f[:-2]:
      is_subfont = True
      subfont_idx = int(f[-2:], 16)
      if cur_font == n and cur_ssize == z:
        return False
      else:
        cur_font = n; cur_ssize = z
        return True
  else:
    is_subfont = False
    cur_font = f; cur_ssize = z
    return True

############################################################
# DVI class
############################################################
class DVI(object):
  def __init__(self, unit='pt'):
    if   unit == 'sp': self.byconv = self.by_sp_conv
    elif unit == 'bp': self.byconv = self.by_bp_conv
    elif unit == 'mm': self.byconv = self.by_mm_conv
    elif unit == 'cm': self.byconv = self.by_cm_conv
    elif unit == 'in': self.byconv = self.by_in_conv
    else:              self.byconv = self.by_pt_conv
    self.Initialize()

  ##########################################################
  # Initialize: Required by __init__(), Load(), and Parse()
  ##########################################################
  def Initialize(self):
    self.id = DVI_ID
    self.numerator   = 25400000
    self.denominator = 473628672
    self.mag = 1000
    self.ComputeConversionFactors()
    self.comment = ''
    self.font_def = {}
    self.max_v = self.max_h = self.max_s = self.total_pages = 0
    self.pages = []

  ##########################################################
  # Load: DVI -> Internal Format
  ##########################################################
  def Load(self, fn):
    fp = open(fn, 'rb')
    self.LoadFromFile(fp)
    fp.close()

  def LoadFromFile(self, fp):
    self.Initialize()
    fp.seek(0, 2)
    if fp.tell() < 53: BadDVI('less than 53 bytes long')
    self.ProcessPreamble(fp)
    self.ProcessPostamble(fp)
    loc = self.first_backpointer
    while loc >= 0:
      fp.seek(loc)
      if GetByte(fp) != BOP: BadDVI('byte %d is not bop' % fp.tell())
      cnt = [SignedQuad(fp) for i in range(10)]
      loc = SignedQuad(fp)
      page = self.ProcessPage(fp)
      self.pages.insert(0, {'count':cnt, 'content':page})

  def ProcessPreamble(self, fp):
    fp.seek(0)
    if GetByte(fp) != PRE: BadDVI("First byte isn't start of preamble")
    id = GetByte(fp)
    if not ValidID(id):
      warning("ID byte is %d; use the default %d!" % (id, DVI_ID))
    self.id = id
    numerator = SignedQuad(fp)
    if numerator <= 0:
      warning('numerator is %d; use the default 25400000!' % numerator)
    else:
      self.numerator = numerator
    denominator = SignedQuad(fp)
    if denominator <= 0:
      warning('denominator is %d; use the default 473628672!' % denominator)
    else:
      self.denominator = denominator
    mag = SignedQuad(fp)
    if mag <= 0:
      warning('magnification is %d; use the default 1000!' % mag)
    else:
      self.mag = mag
    self.comment = fp.read(GetByte(fp)).decode('utf8')
    self.ComputeConversionFactors()

  def ProcessPostamble(self, fp):
    fp.seek(-5, 2) # at least four 223's
    while True:
      k = GetByte(fp)
      if   k < 0:    BadDVI('all 223s; is it a DVI file?') # found EOF
      elif k != 223: break
      fp.seek(-2, 1)
    if not ValidID(k):
      warning('ID byte is %d' % k)
    fp.seek(-5, 1)
    q = SignedQuad(fp)
    m = fp.tell() # id_byte
    if q < 0 or q > m - 33: BadDVI('post pointer %d at byte %d' % (q, m - 4))
    fp.seek(q) # move to post
    k = GetByte(fp)
    if k != POST: BadDVI('byte %d is not post' % k)
    self.post_loc = q
    self.first_backpointer = SignedQuad(fp)

    if SignedQuad(fp) != self.numerator:
      warning("numerator doesn't match the preamble!")
    if SignedQuad(fp) != self.denominator:
      warning("denominator doesn't match the preamble!")
    if SignedQuad(fp) != self.mag:
      warning("magnification doesn't match the preamble!")
    self.max_v = SignedQuad(fp)
    self.max_h = SignedQuad(fp)
    self.max_s = Get2Bytes(fp)
    self.total_pages = Get2Bytes(fp)
    while True:
      k = GetByte(fp)
      if   k == FNT_DEF1: p = GetByte(fp)
      elif k == FNT_DEF2: p = Get2Bytes(fp)
      elif k == FNT_DEF3: p = Get3Bytes(fp)
      elif k == FNT_DEF4: p = SignedQuad(fp)
      elif k == NATIVE_FONT_DEF: p = SignedQuad(fp)
      elif k != NOP: break
      if k == NATIVE_FONT_DEF: self.DefineNativeFont(p, fp)
      else: self.DefineFont(p, fp)
    if k != POST_POST:
      warning('byte %d is not postpost!' % (fp.tell() - 1))
    if SignedQuad(fp) != self.post_loc:
      warning('bad postamble pointer in byte %d!' % (fp.tell() - 4))
    m = GetByte(fp)
    if not ValidID(m):
      warning('identification in byte %d should be one of: %s!' % (fp.tell() - 1, DVI_IDS))
    if not self.id == m:
      if not (self.id == 2 and m == 3): # pTeX/upTeX with dir allowed
        warning('ID byte mismatch: preamble %d vs postamble %d!' % (self.id, m))

  def DefineFont(self, e, fp):
    c = SignedQuad(fp) # font_check_sum
    q = SignedQuad(fp) # font_scaled_size
    d = SignedQuad(fp) # font_design_size
    n = fp.read(GetByte(fp) + GetByte(fp)).decode('utf8')
    try:
      f = self.font_def[e]
    except KeyError:
      self.font_def[e] = {'name':n, 'checksum':c, 'scaled_size':q, 'design_size':d}
      if q <= 0 or q >= 0o1000000000:
        warning("%s---not loaded, bad scale (%d)!" % (n, q))
      elif d <= 0 or d >= 0o1000000000:
        warning("%s---not loaded, bad design size (%d)!" % (n, d))
    else:
      if f['checksum'] != c:
        warning("\t---check sum doesn't match previous definition!")
      if f['scaled_size'] != q:
        warning("\t---scaled size doesn't match previous definition!")
      if f['design_size'] != d:
        warning("\t---design size doesn't match previous definition!")
      if f['name'] != n:
        warning("\t---font name doesn't match previous definition!")

  def DefineNativeFont(self, e, fp):
    size = Get4Bytes(fp) # scaled size
    flags = Get2Bytes(fp)
    l = GetByte(fp) # name length
    fnt_name = fp.read(l).decode('utf8')
    index = Get4Bytes(fp) # face index
    ext = []
    embolden = 0
    if flags & XDV_FLAG_VERTICAL: ext.append("vertical")
    if flags & XDV_FLAG_COLORED: ext.append("color=%08X" % Get4Bytes(fp))
    if flags & XDV_FLAG_EXTEND: ext.append("extend=%d" % SignedQuad(fp))
    if flags & XDV_FLAG_SLANT: ext.append("slant=%d" % SignedQuad(fp))
    if flags & XDV_FLAG_EMBOLDEN: ext.append("embolden=%d" % SignedQuad(fp))
    try:
      f = self.font_def[e]
    except KeyError:
      if index > 0:
        fnt_name += "[%d]" % index
      name = '"%s"' % fnt_name
      if ext:
        name = '"%s:%s"' % (fnt_name, ";".join(ext))

      self.font_def[e] = {
        'name': name,
        'checksum': 0,
        'scaled_size': size,
        'design_size': 655360, # hardcoded
        }

  def ProcessPage(self, fp):
    s = []
    while True:
      o = GetByte(fp)
      p = self.Get1Arg(o, fp)
      if o < SET_CHAR_0 + 128 or o in (SET1, SET2, SET3, SET4):
        q = [p]
        while True:
          o = GetByte(fp)
          p = self.Get1Arg(o, fp)
          if o < SET_CHAR_0 + 128 or o in (SET1, SET2, SET3, SET4):
            q.append(p)
          else:
            break
        s.append([SET1, q])
      if o == SET_RULE:
        s.append([SET_RULE, [p, SignedQuad(fp)]])
      elif o in (PUT1, PUT2, PUT3, PUT4):
        s.append([PUT1, [p]])
      elif o == PUT_RULE:
        s.append([PUT_RULE, [p, SignedQuad(fp)]])
      elif o == NOP:
        continue
      elif o == BOP:
        warning('bop occurred before eop!')
        break
      elif o == EOP:
        break
      elif o == PUSH:
        s.append([PUSH])
      elif o == POP:
        s.append([POP])
      elif o in (RIGHT1, RIGHT2, RIGHT3, RIGHT4):
        s.append([RIGHT1, p])
      elif o == W0:
        s.append([W0])
      elif o in (W1, W2, W3, W4):
        s.append([W1, p])
      elif o == X0:
        s.append([X0])
      elif o in (X1, X2, X3, X4):
        s.append([X1, p])
      elif o in (DOWN1, DOWN2, DOWN3, DOWN4):
        s.append([DOWN1, p])
      elif o == Y0:
        s.append([Y0])
      elif o in (Y1, Y2, Y3, Y4):
        s.append([Y1, p])
      elif o == Z0:
        s.append([Z0])
      elif o in (Z1, Z2, Z3, Z4):
        s.append([Z1, p])
      elif o < FNT_NUM_0 + 64 or o in (FNT1, FNT2, FNT3, FNT4):
        s.append([FNT1, p])
      elif o in (XXX1, XXX2, XXX3, XXX4):
        q = fp.read(p)
        s.append([XXX1, q])
      elif o in (FNT_DEF1, FNT_DEF2, FNT_DEF3, FNT_DEF4):
        self.DefineFont(p, fp)
      elif o == NATIVE_FONT_DEF:
        self.DefineNativeFont(p, fp)
      elif o == GLYPHS:
        s.append([GLYPHS, self.GetGlyphs(fp)])
      elif o == TEXT_GLYPHS:
        s.append([TEXT_GLYPHS, self.GetTextGlyphs(fp)])
      elif o == DIR:
        s.append([DIR, p])
      elif o == BEGIN_REFLECT:
        s.append([BEGIN_REFLECT])
      elif o == END_REFLECT:
        s.append([END_REFLECT])
      elif o == PRE:
        warning('preamble command within a page!')
        break
      elif o in (POST, POST_POST):
        warning('postamble command %d!' % o)
        break
      else:
        warning('undefined command %d!' % o)
        break
    return s

  def Get1Arg(self, o, fp):
    if o < SET_CHAR_0 + 128:
      return o - SET_CHAR_0
    if o in (SET1, PUT1, FNT1, XXX1, FNT_DEF1, DIR):
      return GetByte(fp)
    if o in (SET2, PUT2, FNT2, XXX2, FNT_DEF2):
      return Get2Bytes(fp)
    if o in (SET3, PUT3, FNT3, XXX3, FNT_DEF3):
      return Get3Bytes(fp)
    if o in (RIGHT1, W1, X1, DOWN1, Y1, Z1):
      return SignedByte(fp)
    if o in (RIGHT2, W2, X2, DOWN2, Y2, Z2):
      return SignedPair(fp)
    if o in (RIGHT3, W3, X3, DOWN3, Y3, Z3):
      return SignedTrio(fp)
    if o in (SET4, SET_RULE, PUT4, PUT_RULE, RIGHT4, W4, X4, DOWN4, Y4, Z4, FNT4, XXX4, FNT_DEF4, NATIVE_FONT_DEF):
      return SignedQuad(fp)
    if o in (NOP, BOP, EOP, PUSH, POP, PRE, POST, POST_POST) or o > POST_POST:
      return 0
    if o in (W0, X0, Y0, Z0, BEGIN_REFLECT, END_REFLECT):
      return 0
    if o < FNT_NUM_0 + 64:
      return o - FNT_NUM_0

  def GetGlyphs(self, fp):
    width = SignedQuad(fp)
    length = Get2Bytes(fp)
    glyphs = {}
    for i in range(length):
      glyphs[i] = {}
      glyphs[i]["x"] = SignedQuad(fp)
      glyphs[i]["y"] = SignedQuad(fp)

    for i in range(length):
      glyphs[i]["id"] = Get2Bytes(fp)

    return (width, glyphs)

  def GetTextGlyphs(self, fp):
    length = Get2Bytes(fp)
    chars = []
    for i in range(length):
      chars.append(Get2Bytes(fp))
    width, glyphs = self.GetGlyphs(fp)

    return (chars, width, glyphs)

  def ReadGlyphs(self, val):
    import re
    glyphs = []
    w, g = val.split(" ", 1)
    for m in re.finditer(r"gid(?P<id>\d+?)\((?P<pos>.*?.)\)", g):
      gid = m.group("id")
      pos = m.group("pos")

      if "," in pos:
        x, y = pos.split(",")
      else:
        x, y = pos, "0sp"

      glyphs.append({"id": int(gid), 'x': self.ConvLen(x), 'y': self.ConvLen(y)})

    return (self.ConvLen(w), glyphs)

  def ReadTextGlyphs(self, val):
    _, text, glyphs = val.split(val[0])
    text = "'%s'" % text
    glyphs = glyphs.lstrip()
    chars = GetStr(text)
    w, glyphs = self.ReadGlyphs(glyphs)

    return (chars, w, glyphs)

  ##########################################################
  # Save: Internal Format -> DVI
  ##########################################################
  def Save(self, fn):
    fp = open(fn, 'wb')
    self.SaveToFile(fp)
    fp.close()

  def SaveToFile(self, fp):
    # WritePreamble
    fp.write(b''.join([bytes.fromhex('%02x' % PRE), PutByte(self.id), PutSignedQuad(self.numerator), PutSignedQuad(self.denominator), PutSignedQuad(self.mag), PutByte(len(self.comment)), self.comment.encode('utf8')]))
    # WriteFontDefinitions
    self.WriteFontDefinitions(fp)
    # WritePages
    if not self.pages:
      # dvistd0.pdf, Section A.1:
      # > A DVI file consists of a ``preamble,'' followed by a sequence of
      # > one or more "pages," followed by a ``postamble.''
      warning('one or more pages required!')
      self.pages.append({'count':[1,0,0,0,0,0,0,0,0,0], 'content':[]})
    stackdepth = 0; loc = -1
    for page in self.pages:
      w = x = y = z = 0; stack = []
      s = [bytes.fromhex('%02x' % BOP)]
      s.extend([PutSignedQuad(c) for c in page['count']])
      s.append(PutSignedQuad(loc))
      for cmd in page['content']:
        if cmd[0] == SET1:
          for o in cmd[1]:
            if o < 128: s.append(bytes.fromhex('%02x' % (SET_CHAR_0 + o)))
            else:       s.append(self.CmdPairU([SET1, o]))
        elif cmd[0] in (SET_RULE, PUT_RULE):
          s.append(bytes.fromhex('%02x' % cmd[0]) + PutSignedQuad(cmd[1][0]) + PutSignedQuad(cmd[1][1]))
        elif cmd[0] == PUT1:
          s.append(self.CmdPairU([PUT1, cmd[1][0]]))
        elif cmd[0] in (RIGHT1, DOWN1):
          s.append(self.CmdPair(cmd))
        elif cmd[0] in (W0, X0, Y0, Z0):
          s.append(bytes.fromhex('%02x' % cmd[0]))
        elif cmd[0] == PUSH:
          s.append(bytes.fromhex('%02x' % PUSH))
          stack.append((w, x, y, z))
          if len(stack) > stackdepth: stackdepth = len(stack)
        elif cmd[0] == POP:
          s.append(bytes.fromhex('%02x' % POP))
          w, x, y, z = stack.pop()
        elif cmd[0] == W1:
          w = cmd[1]; s.append(self.CmdPair(cmd))
        elif cmd[0] == X1:
          x = cmd[1]; s.append(self.CmdPair(cmd))
        elif cmd[0] == Y1:
          y = cmd[1]; s.append(self.CmdPair(cmd))
        elif cmd[0] == Z1:
          z = cmd[1]; s.append(self.CmdPair(cmd))
        elif cmd[0] == FNT1:
          if cmd[1] < 64: s.append(bytes.fromhex('%02x' % (FNT_NUM_0 + cmd[1])))
          else:           s.append(self.CmdPairU(cmd))
        elif cmd[0] == XXX1:
          if options.xxx_encoding == "none":
            l = len(cmd[1]) # leave encoding untouched
          else:
            cmd1 = cmd[1].encode(options.xxx_encoding)
            l = len(cmd1)
          if l < 256:
            s.append(bytes.fromhex('%02x' % XXX1) + bytes.fromhex('%02x' % l))
          else:
            s.append(bytes.fromhex('%02x' % XXX4) + PutSignedQuad(l))
          if options.xxx_encoding == "none":
            for o in cmd[1]:
              s.append(bytes.fromhex('%02x' % ord(o)))
          else:
              s.append(cmd1)
        elif cmd[0] == DIR:
          s.append(bytes.fromhex('%02x' % DIR) + bytes.fromhex('%02x' % cmd[1]))
        elif cmd[0] == BEGIN_REFLECT:
          s.append(bytes.fromhex('%02x' % BEGIN_REFLECT))
        elif cmd[0] == END_REFLECT:
          s.append(bytes.fromhex('%02x' % END_REFLECT))
        elif cmd[0] == GLYPHS:
          s.append(PutByte(GLYPHS))
          s.append(PutGlyphs(cmd[1], cmd[2]))
        elif cmd[0] == TEXT_GLYPHS:
          s.append(PutByte(TEXT_GLYPHS))
          s.append(PutTextGlyphs(cmd[1], cmd[2], cmd[3]))
        else:
          warning('invalid command %s!' % cmd[0])
      s.append(bytes.fromhex('%02x' % EOP))
      loc = fp.tell()
      fp.write(b''.join(s))
    # WritePostamble
    post_loc = fp.tell()
    fp.write(b''.join([bytes.fromhex('%02x' % POST), PutSignedQuad(loc), PutSignedQuad(self.numerator), PutSignedQuad(self.denominator), PutSignedQuad(self.mag), PutSignedQuad(self.max_v), PutSignedQuad(self.max_h), Put2Bytes(stackdepth), Put2Bytes(len(self.pages))]))
    # WriteFontDefinitions
    self.WriteFontDefinitions(fp)
    # WritePostPostamble
    fp.write(b''.join([bytes.fromhex('%02x' % POST_POST), PutSignedQuad(post_loc), PutByte(self.id_post), b'\xdf\xdf\xdf\xdf']))
    loc = fp.tell()
    while (loc % 4) != 0:
      fp.write(b'\xdf'); loc += 1

  def WriteFontDefinitions(self, fp):
    s = []
    for e in sorted(self.font_def.keys()):
      try:
        self.font_def[e]['native']
        flags = self.font_def[e]['flags']
        s.append(PutByte(NATIVE_FONT_DEF))
        s.append(PutSignedQuad(e))
        s.append(PutSignedQuad(self.font_def[e]['scaled_size']))
        s.append(Put2Bytes(flags))
        s.append(PutByte(len(self.font_def[e]['name'])))
        s.append(self.font_def[e]['name'].encode('utf8'))
        s.append(PutSignedQuad(self.font_def[e]['index']))
        if flags & XDV_FLAG_COLORED: s.append(PutSignedQuad(self.font_def[e]['color']))
        if flags & XDV_FLAG_EXTEND: s.append(PutSignedQuad(self.font_def[e]['extend']))
        if flags & XDV_FLAG_SLANT: s.append(PutSignedQuad(self.font_def[e]['slant']))
        if flags & XDV_FLAG_EMBOLDEN: s.append(PutSignedQuad(self.font_def[e]['embolden']))
      except KeyError:
        l, q = PutUnsigned(e)
        s.append(PutByte(FNT_DEF1 + l))
        s.append(q)
        s.append(PutSignedQuad(self.font_def[e]['checksum']))
        s.append(PutSignedQuad(self.font_def[e]['scaled_size']))
        s.append(PutSignedQuad(self.font_def[e]['design_size']))
        s.append(b'\x00')
        s.append(PutByte(len(self.font_def[e]['name'])))
        s.append(self.font_def[e]['name'].encode('utf8'))
    fp.write(b''.join(s))

  def CmdPairU(self, cmd):
    l, q = PutUnsigned(cmd[1])
    return bytes.fromhex('%02x' % (cmd[0] + l)) + q

  def CmdPair(self, cmd):
    l, q = PutSigned(cmd[1])
    return bytes.fromhex('%02x' % (cmd[0] + l)) + q

  ##########################################################
  # Parse: Text -> Internal Format
  ##########################################################
  def Parse(self, fn, encoding=''):
    fp = open(fn, 'r', encoding=encoding)
    s = fp.read()
    fp.close()
    self.ParseFromString(s, encoding=encoding)

  def ParseFromString(self, s, encoding=''):
    global GetStr, cur_font, cur_dsize, cur_ssize, subfont_idx
    if encoding == 'ascii': GetStr = GetStrASCII
    else:                   GetStr = GetStrUTF8
    self.Initialize()
    self.fnt_num = 0
    dir_used = 0
    for l in s.split('\n'):
      l = l.strip()
      if not l or l[0] == '%': continue
      try:
        key, val = l.split(':', 1)
        key = key.strip(); val = val.strip()
      except:
        if l[-1] == ']': v = l[:-1].split(' ')
        else: v = l.split(' ')
        if v[0] == "[page":
          self.cur_page = []
          count = [GetInt(c) for c in v[1:]]
          if len(count) < 10: count += ([0] * (10-len(count)))
          self.pages.append({'count':count, 'content':self.cur_page})
        continue
      # ParsePreamble
      if key == "id":
        self.id = GetInt(val)
        if not ValidID(self.id):
          warning('identification byte %d should be one of: %s!' % (self.id, DVI_IDS))
      elif key == "numerator":
        d = GetInt(val)
        if d <= 0:
          warning('non-positive numerator %d!' % d)
        else:
          self.numerator = d
          self.ComputeConversionFactors()
      elif key == "denominator":
        d = GetInt(val)
        if d <= 0:
          warning('non-positive denominator %d!' % d)
        else:
          self.denominator = d
          self.ComputeConversionFactors()
      elif key == "magnification":
        d = GetInt(val)
        if d <= 0:
          warning('non-positive magnification %d!' % d)
        else:
          self.mag = d
      elif key == "comment":
        self.comment = val[1:-1]
      # Parse Postamble
      elif key == "maxv":
        self.max_v = self.ConvLen(val)
      elif key == "maxh":
        self.max_h = self.ConvLen(val)
      elif key == "maxs":
        self.max_s = GetInt(val)
      elif key == "pages":
        self.total_pages = GetInt(val)
      # Parse Font Definitions
      elif key == "fntdef":
        self.font_def[self.fnt_num] = self.GetFntDef(val)
        self.fnt_num += 1
      # Parse Pages
      elif key == 'xxx':
        self.cur_page.append([XXX1, eval(val)])
      elif key == 'set':
        ol = GetStr(val)
        if is_subfont:
          subfont_idx = (ol[0] >> 8)
          self.AppendFNT1()
          nl = [ol[0] & 0xff]
          for o in ol[1:]:
            idx = (o >> 8)
            if idx != subfont_idx:
              self.cur_page.append([SET1, nl])
              subfont_idx = idx
              self.AppendFNT1()
              nl = [o & 0xff]
            else:
              nl.append(o & 0xff)
          self.cur_page.append([SET1, nl])
        else:
          self.cur_page.append([SET1, ol])
      elif key == 'put':
        ol = GetStr(val)
        if len(ol) != 1:
          warning('only one character is allowed for put!')
        self.cur_page.append([PUT1, ol])
      elif key == 'setrule':
        v = val.split(' ')
        if len(v) != 2:
          warning('two values are required for setrule!')
          continue
        self.cur_page.append([SET_RULE, [self.ConvLen(c) for c in v]])
      elif key == 'putrule':
        v = val.split(' ')
        if len(v) != 2:
          warning('two values are required for putrule!')
          continue
        self.cur_page.append([PUT_RULE, [self.ConvLen(c) for c in v]])
      elif key == 'fnt':
        f = self.GetFntDef(val)
        n = f['name']
        d = f['design_size']
        q = f['scaled_size']
        if n in subfont_list:
          is_subfont = True
          cur_font = n; cur_dsize = d; cur_ssize = q
        else:
          is_subfont = False
          try:
            e = list(self.font_def.keys())[list(self.font_def.values()).index(f)]
          except:
            e = self.fnt_num
            self.font_def[self.fnt_num] = f
            self.fnt_num += 1
          self.cur_page.append([FNT1, e])
      elif key == 'right':
        self.cur_page.append([RIGHT1, self.ConvLen(val)])
      elif key == 'down':
        self.cur_page.append([DOWN1, self.ConvLen(val)])
      elif key == 'w':
        self.cur_page.append([W1, self.ConvLen(val)])
      elif key == 'x':
        self.cur_page.append([X1, self.ConvLen(val)])
      elif key == 'y':
        self.cur_page.append([Y1, self.ConvLen(val)])
      elif key == 'z':
        self.cur_page.append([Z1, self.ConvLen(val)])
      elif key == 'push':
        self.cur_page.append([PUSH])
      elif key == 'pop':
        self.cur_page.append([POP])
      elif key == 'w0':
        self.cur_page.append([W0])
      elif key == 'x0':
        self.cur_page.append([X0])
      elif key == 'y0':
        self.cur_page.append([Y0])
      elif key == 'z0':
        self.cur_page.append([Z0])
      elif key == 'dir':
        self.cur_page.append([DIR, GetInt(val)])
        dir_used = 1
      elif key == 'begin_reflect':
        self.cur_page.append([BEGIN_REFLECT])
      elif key == 'end_reflect':
        self.cur_page.append([END_REFLECT])
      elif key == 'setglyphs':
        w, glyphs = self.ReadGlyphs(val)
        self.cur_page.append([GLYPHS, w, glyphs])
      elif key == 'settextglyphs':
        text, w, glyphs = self.ReadTextGlyphs(val)
        self.cur_page.append([TEXT_GLYPHS, text, w, glyphs])
      else:
        warning('invalid command %s!' % key)
    if self.id == 2 and dir_used == 1: # standard DVI with dir -> force pTeX/upTeX spec
      self.id_post = 3
    else:
      self.id_post = self.id

  def AppendFNT1(self):
    f = {'name':cur_font+"%02x"%subfont_idx, 'design_size':cur_dsize, 'scaled_size':cur_ssize, 'checksum':0}
    try:
      e = list(self.font_def.keys())[list(self.font_def.values()).index(f)]
    except:
      e = self.fnt_num
      self.font_def[e] = f
      self.fnt_num += 1
    self.cur_page.append([FNT1, e])

  ##########################################################
  # Dump: Internal Format -> Text
  ##########################################################
  def Dump(self, fn, tabsize=2, encoding=''):
    fp = open(fn, 'w', encoding=encoding)
    self.DumpToFile(fp, tabsize=tabsize, encoding=encoding)
    fp.close()

  def DumpToFile(self, fp, tabsize=2, encoding=''):
    global PutStr
    if   encoding == 'ascii':  PutStr = PutStrASCII
    elif encoding == 'latin1': PutStr = PutStrLatin1
    else:                      PutStr = PutStrUTF8
    # DumpPreamble
    fp.write("[preamble]\n")
    fp.write("id: %d\n" % self.id)
    fp.write("numerator: %d\n" % self.numerator)
    fp.write("denominator: %d\n" % self.denominator)
    fp.write("magnification: %d\n" % self.mag)
    fp.write("comment: %s\n" % repr(self.comment))
    # DumpPostamble
    fp.write("\n[postamble]\n")
    fp.write("maxv: %s\n" % self.byconv(self.max_v))
    fp.write("maxh: %s\n" % self.byconv(self.max_h))
    fp.write("maxs: %d\n" % self.max_s)
    fp.write("pages: %d\n" % self.total_pages)
    # DumpFontDefinitions
    fp.write("\n[font definitions]\n")
    for e in sorted(self.font_def.keys()):
      fp.write("fntdef: %s " % self.font_def[e]['name'])
      if self.font_def[e]['design_size'] != self.font_def[e]['scaled_size']:
        fp.write("(%s) " % self.byconv(self.font_def[e]['design_size']))
      fp.write("at %s\n" % self.byconv(self.font_def[e]['scaled_size']))
    # DumpPages
    if not self.pages:
      # dvistd0.pdf, Section A.1:
      # > A DVI file consists of a ``preamble,'' followed by a sequence of
      # > one or more "pages," followed by a ``postamble.''
      warning('one or more pages required!')
      self.pages.append({'count':[1,0,0,0,0,0,0,0,0,0], 'content':[]})
    for page in self.pages:
      fp.write("\n[page" + (" %d"*10 % tuple(page['count'])) + "]\n")
      indent = 0
      for cmd in page['content']:
        if cmd[0] == POP:
          indent -= tabsize
          fp.write("%spop:\n" % (' ' * indent))
          continue
        fp.write("%s" % (' ' * indent))
        if cmd[0] == PUSH:
          fp.write("push:\n")
          indent += tabsize
        elif cmd[0] == XXX1:
          if options.xxx_encoding == "none":
            fp.write("xxx: %s\n" % PutStrASCII(cmd[1])) # leave encoding untouched
          else:
            fp.write("xxx: '%s'\n" % cmd[1].decode(options.xxx_encoding))
        elif cmd[0] == DIR:
          fp.write("dir: %d\n" % cmd[1])
        elif cmd[0] == BEGIN_REFLECT:
          fp.write("begin_reflect:\n")
        elif cmd[0] == END_REFLECT:
          fp.write("end_reflect:\n")
        elif cmd[0] == SET_RULE:
          fp.write("setrule: %s %s\n" % (self.byconv(cmd[1][0]), self.byconv(cmd[1][1])))
        elif cmd[0] == PUT_RULE:
          fp.write("putrule: %s %s\n" % (self.byconv(cmd[1][0]), self.byconv(cmd[1][1])))
        elif cmd[0] == SET1:
          fp.write("set: %s\n" % PutStr(cmd[1]))
        elif cmd[0] == PUT1:
          fp.write("put: %s\n" % PutStr(cmd[1]))
        elif cmd[0] == FNT1:
          f = self.font_def[cmd[1]]['name']
          z = self.font_def[cmd[1]]['scaled_size']
          if IsFontChanged(f, z):
            fp.write("fnt: %s " % cur_font)
            if self.font_def[cmd[1]]['design_size'] != self.font_def[cmd[1]]['scaled_size']:
              fp.write("(%s) " % self.byconv(self.font_def[cmd[1]]['design_size']))
            fp.write("at %s\n" % self.byconv(cur_ssize))
        elif cmd[0] == GLYPHS:
          fp.write("setglyphs: %s\n" % self.DumpGlyphs(cmd[1][0], cmd[1][1]))
        elif cmd[0] == TEXT_GLYPHS:
          fp.write("settextglyphs: %s\n" % self.DumpTextGlyphs(cmd[1][0], cmd[1][1], cmd[1][2]))
        elif cmd[0] == RIGHT1:
          fp.write("right: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == DOWN1:
          fp.write("down: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == W1:
          fp.write("w: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == X1:
          fp.write("x: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == Y1:
          fp.write("y: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == Z1:
          fp.write("z: %s\n" % self.byconv(cmd[1]))
        elif cmd[0] == W0:
          fp.write("w0:\n")
        elif cmd[0] == X0:
          fp.write("x0:\n")
        elif cmd[0] == Y0:
          fp.write("y0:\n")
        elif cmd[0] == Z0:
          fp.write("z0:\n")

  def DumpGlyphs(self, w, g):
    yPresent = False
    for i in g:
      if g[i]["y"] != 0:
        yPresent = True

    glyphs = []
    for i in g:
      gid = "gid%s" % g[i]["id"]
      x = self.byconv(g[i]["x"])
      y = self.byconv(g[i]["y"])
      if yPresent:
        glyphs.append("%s(%s, %s)" % (gid, x, y))
      else:
        glyphs.append("%s(%s)" % (gid, x))

    return "%s %s" % (self.byconv(w), " ".join(glyphs))

  def DumpTextGlyphs(self, t, w, g):
    return "%s %s" % (PutStrUTF8(t), self.DumpGlyphs(w, g))

  ##########################################################
  # Misc Functions
  ##########################################################
  def ComputeConversionFactors(self):
    self.sp_conv = (self.numerator / 25400000.) * (473628672. / self.denominator)
    self.pt_conv = (self.numerator / 25400000.) * (7227. / self.denominator)
    self.bp_conv = (self.numerator / 254000.) * (72. / self.denominator)
    self.mm_conv = (self.numerator / 10000.) / self.denominator
    self.cm_conv = (self.numerator / 100000.) / self.denominator
    self.in_conv = (self.numerator / 254000.) * (1. / self.denominator)

  def ConvLen(self, s):
    try:    return int(s)
    except: pass
    try:    f = float(s[:-2])
    except: return 0
    m = s[-2:]
    if   m == "pt": return int(round(f / self.pt_conv))
    elif m == "in": return int(round(f / self.in_conv))
    elif m == "mm": return int(round(f / self.mm_conv))
    elif m == "cm": return int(round(f / self.cm_conv))
    elif m == "bp": return int(round(f / self.bp_conv))
    elif m == "sp": return int(round(f / self.sp_conv))
    else:
      try:    return int(round(f / self.pt_conv))
      except: return 0

  def GetFntDef(self, s):
    f = {}
    try:
      n, size = s.split('(', 1)
      d, q = size.split(')', 1)
    except:
      n, q = s.split(' ', 1)
    n = n.strip(); q = q.strip()
    if n.startswith('"') and n.endswith('"'):
      f['native'] = True
      n = n.strip('"')
      flags = 0
      color = 0
      extend = 0
      slant = 0
      embolden = 0
      try:
        name, ext = n.split(':')
      except:
        name, ext = n, ""

      try:
        name, index = name.split('[')
        index = index.split(']')[0]
      except:
        index = 0

      if ext:
        ext = ext.split(';')
        for opt in ext:
          try:
            key, value = opt.split('=')
          except:
            key, value = opt, ""
          if key == "color":
            flags |= XDV_FLAG_COLORED
            color = int(value, 16)
          if key == "vertical":
            flags |= XDV_FLAG_VERTICAL
          if key == "extend":
            flags |= XDV_FLAG_EXTEND
            extend = int(value)
          if key == "slant":
            flags |= XDV_FLAG_SLANT
            slant = int(value)
          if key == "embolden":
            flags |= XDV_FLAG_EMBOLDEN
            embolden = int(value)

      f['name'] = name
      f['index'] = int(index)
      f['flags'] = flags
      f['color'] = color
      f['extend'] = extend
      f['slant'] = slant
      f['embolden'] = embolden
    else:
      f['name'] = n

    if q[:2] == "at": q = q[2:]
    q = self.ConvLen(q.strip())
    try:    d = self.ConvLen(d.strip())
    except: d = q

    f['design_size'] = d
    f['scaled_size'] = q
    f['checksum'] = 0

    return f

  def by_sp_conv(self, a):
    v = self.sp_conv * a
    return "%dsp" % int(v)

  def by_pt_conv(self, a):
    v = self.pt_conv * a
    if v == int(v): return "%dpt" % int(v)
    else:           return "%fpt" % v

  def by_bp_conv(self, a):
    v = self.bp_conv * a
    if v == int(v): return "%dbp" % int(v)
    else:           return "%fbp" % v

  def by_mm_conv(self, a):
    v = self.mm_conv * a
    if v == int(v): return "%dmm" % int(v)
    else:           return "%fmm" % v

  def by_cm_conv(self, a):
    v = self.cm_conv * a
    if v == int(v): return "%dcm" % int(v)
    else:           return "%fcm" % v

  def by_in_conv(self, a):
    v = self.in_conv * a
    if v == int(v): return "%din" % int(v)
    else:           return "%fin" % v

############################################################
# Misc Functions for Main Routine
############################################################
def ProcessOptions():
  usage = """%prog [options] dvi_file|dvi_dump_file

DVIasm is a Python script to support changing or creating DVI files
via disassembling into text, editing, and then reassembling into
binary format. It is fully documented at
  https://tug.org/TUGboat/Articles/tb28-2/tb89cho.pdf
  http://ajt.ktug.kr/assets/2008/5/1/0201cho.pdf

Please report bugs to
  https://github.com/aminophen/dviasm"""

  version = """This is %prog-20240725

Copyright (C) 2007-2008 by Jin-Hwan Cho <chofchof@ktug.or.kr>
Copyright (C) 2011-2017 by Khaled Hosny <khaledhosny@eglug.org>
Copyright (C) 2019      by Arthur Reutenauer <arthur@reutenauer.eu>
Copyright (C) 2019-2024 by Hironobu Yamashita <h.y.acetaminophen@gmail.com>

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version."""

  parser = OptionParser(usage=usage, version=version)
  parser.add_option("-u", "--unit",
                    action="store", type="string", dest="unit",
                    metavar="STR",
                    help="unit (sp, pt, bp, mm, cm, in) [default=%default]")
  parser.add_option("-o", "--output",
                    action="store", type="string", dest="output",
                    metavar="FILE",
                    help="filename for output instead of stdout")
  parser.add_option("-e", "--encoding",
                    action="store", type="string", dest="encoding",
                    metavar="STR",
                    help="encoding for input/output [default=%default]")
  parser.add_option("-x", "--xxx-encoding",
                    action="store", type="string", dest="xxx_encoding",
                    metavar="STR",
                    help="encoding for interpreting xxx strings [default=%default]")
  parser.add_option("-t", "--tabsize",
                    action="store", type="int", dest="tabsize",
                    metavar="INT",
                    help="tab size for push/pop [default=%default]")
  parser.add_option("-p", "--ptex",
                    action="store_true", dest="ptex", default=False,
                    help="ISO-2022-JP-encoded DVI for Japanese pTeX")
  parser.add_option("-s", "--subfont",
                    action="append", type="string", dest="subfont",
                    metavar="STR",
                    help="the list of fonts with UCS2 subfont scheme (comma separated); disable internal subfont list if STR is empty")
  parser.set_defaults(unit='pt', encoding='utf8', xxx_encoding='none', tabsize=2)
  (options, args) = parser.parse_args()
  if not options.unit in ['sp', 'pt', 'bp', 'mm', 'cm', 'in']:
    parser.error("invalid unit name '%s'!" % options.unit)
  if options.tabsize < 0:
    parser.error("negative tabsize!")
  if not options.xxx_encoding in ['none', 'utf8', 'sjis', 'eucjp']:
    parser.error("invalid xxx-encoding '%s'!" % options.xxx_encoding)
  if not options.encoding in ['ascii', 'latin1', 'utf8', 'sjis', 'eucjp']:
    parser.error("invalid encoding '%s'!" % options.encoding)
  if options.ptex:
    global is_ptex
    is_ptex = True
    if not options.encoding in ['utf8', 'sjis', 'eucjp']:
      parser.error("invalid encoding '%s' for Japanese pTeX!" % options.encoding)
  if options.subfont:
    global subfont_list
    if not options.subfont[0]: # disable subfont
      subfont_list = []
    for l in options.subfont:
      subfont_list.extend([f.strip() for f in l.split(',')])
  if len(args) != 1:
    parser.error("try with the option --help!")
  return (options, args)

def IsDVI(fname):
  from os.path import splitext
  if splitext(fname)[1] not in ('.dvi', '.xdv'): return False
  try:
    fp = open(fname, 'rb')
    fp.seek(0)
    if GetByte(fp) != PRE: return False
    fp.seek(-4, 2)
    if GetByte(fp) != 223: return False
    fp.close()
  except:
    sys.stderr.write('Failed to read %s\n' % fname)
    return False
  return True

############################################################
# Main Routine
############################################################
if __name__ == '__main__':
  (options, args) = ProcessOptions()
  aDVI = DVI(unit=options.unit)
  if os.path.isfile(args[0]): fname = args[0]
  elif os.path.isfile(args[0] + '.xdv'): fname = args[0] + '.xdv'
  elif os.path.isfile(args[0] + '.dvi'): fname = args[0] + '.dvi'
  else:
    sys.stderr.write('File %s not found\n' % args[0])
    sys.exit(1)
  if IsDVI(fname): # dvi -> dump
    aDVI.Load(fname)
    if options.output: aDVI.Dump(options.output, tabsize=options.tabsize, encoding=options.encoding)
    else:              aDVI.DumpToFile(sys.stdout, tabsize=options.tabsize, encoding=options.encoding)
  else: # dump -> dvi
    aDVI.Parse(fname, encoding=options.encoding)
    if options.output: aDVI.Save(options.output)
    else:              aDVI.SaveToFile(sys.stdout.buffer)



#!/usr/bin/env python3

# Python module to feed Asymptote with commands
# (modified from gnuplot.py)
from subprocess import *
class asy:
	def __init__(self):
		self.session = Popen(['asy','-quiet','-inpipe=0','-outpipe=2'],stdin=PIPE)
		self.help()
	def send(self, cmd):
		self.session.stdin.write(bytes(cmd+'\n','utf-8'))
		self.session.stdin.flush()
	def size(self, size):
		self.send("size(%d);" % size)
	def draw(self, str):
		self.send("draw(%s);" % str)
	def fill(self, str):
		self.send("fill(%s);" % str)
	def clip(self, str):
		self.send("clip(%s);" % str)
	def label(self, str):
		self.send("label(%s);" % str)
	def shipout(self, str):
		self.send("shipout(\"%s\");" % str)
	def erase(self):
		self.send("erase();")
	def help(self):
		print("Asymptote session is open.  Available methods are:")
		print("    help(), size(int), draw(str), fill(str), clip(str), label(str), shipout(str), send(str), erase()")
	def __del__(self):
		print("closing Asymptote session...")
		self.send('quit');
		self.session.stdin.close();
		self.session.wait()

if __name__=="__main__":
	g=asy()
	g.size(200)
	g.draw('unitcircle')
	g.send('draw(unitsquare)')
	g.fill('unitsquare,blue')
	g.clip('unitcircle')
	g.label('"$O$",(0,0),SW')
	input('press ENTER to continue')
	g.erase()
	del g



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'labelTextEditor.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDialog,
    QFrame, QGridLayout, QHBoxLayout, QLabel,
    QPlainTextEdit, QPushButton, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)
from xasyicons import icons_rc

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(473, 424)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.StyledPanel)
        self.frame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(-1, 0, -1, -1)
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_2, 0, 2, 1, 1)

        self.chkMathMode = QCheckBox(self.frame)
        self.chkMathMode.setObjectName(u"chkMathMode")

        self.gridLayout.addWidget(self.chkMathMode, 0, 0, 1, 1)

        self.cmbMathStyle = QComboBox(self.frame)
        self.cmbMathStyle.addItem("")
        self.cmbMathStyle.addItem("")
        self.cmbMathStyle.addItem("")
        self.cmbMathStyle.setObjectName(u"cmbMathStyle")
        self.cmbMathStyle.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cmbMathStyle.sizePolicy().hasHeightForWidth())
        self.cmbMathStyle.setSizePolicy(sizePolicy)
        self.cmbMathStyle.setMinimumSize(QSize(100, 0))

        self.gridLayout.addWidget(self.cmbMathStyle, 0, 1, 1, 1)


        self.verticalLayout_3.addLayout(self.gridLayout)

        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.txtLabelEdit = QPlainTextEdit(self.frame)
        self.txtLabelEdit.setObjectName(u"txtLabelEdit")

        self.verticalLayout_2.addWidget(self.txtLabelEdit)


        self.verticalLayout_3.addLayout(self.verticalLayout_2)

        self.verticalLayout_4 = QVBoxLayout()
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(-1, 0, -1, -1)
        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")

        self.verticalLayout_4.addWidget(self.label)

        self.lblLabelPreview = QLabel(self.frame)
        self.lblLabelPreview.setObjectName(u"lblLabelPreview")
        self.lblLabelPreview.setMinimumSize(QSize(0, 100))
        self.lblLabelPreview.setFrameShape(QFrame.Box)

        self.verticalLayout_4.addWidget(self.lblLabelPreview)


        self.verticalLayout_3.addLayout(self.verticalLayout_4)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.btnGetText = QPushButton(self.frame)
        self.btnGetText.setObjectName(u"btnGetText")
        self.btnGetText.setMaximumSize(QSize(32, 32))
        icon = QIcon()
        icon.addFile(u":/icons/text.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.btnGetText.setIcon(icon)
        self.btnGetText.setFlat(True)

        self.horizontalLayout.addWidget(self.btnGetText)

        self.btnPreview = QPushButton(self.frame)
        self.btnPreview.setObjectName(u"btnPreview")
        self.btnPreview.setMaximumSize(QSize(32, 32))
        icon1 = QIcon()
        icon1.addFile(u":/icons/eye.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.btnPreview.setIcon(icon1)
        self.btnPreview.setFlat(True)

        self.horizontalLayout.addWidget(self.btnPreview)

        self.btnCancel = QPushButton(self.frame)
        self.btnCancel.setObjectName(u"btnCancel")
        self.btnCancel.setMaximumSize(QSize(32, 32))
        icon2 = QIcon()
        icon2.addFile(u":/icons/android-close.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.btnCancel.setIcon(icon2)
        self.btnCancel.setFlat(True)

        self.horizontalLayout.addWidget(self.btnCancel)

        self.btnAccept = QPushButton(self.frame)
        self.btnAccept.setObjectName(u"btnAccept")
        self.btnAccept.setMaximumSize(QSize(32, 32))
        icon3 = QIcon()
        icon3.addFile(u":/icons/android-done.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.btnAccept.setIcon(icon3)
        self.btnAccept.setFlat(True)

        self.horizontalLayout.addWidget(self.btnAccept)


        self.verticalLayout_3.addLayout(self.horizontalLayout)


        self.verticalLayout.addWidget(self.frame)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
        self.chkMathMode.setText(QCoreApplication.translate("Dialog", u"Math Mode", None))
        self.cmbMathStyle.setItemText(0, QCoreApplication.translate("Dialog", u"Inline Style", None))
        self.cmbMathStyle.setItemText(1, QCoreApplication.translate("Dialog", u"Display Style", None))
        self.cmbMathStyle.setItemText(2, QCoreApplication.translate("Dialog", u"Script Style", None))

        self.label.setText(QCoreApplication.translate("Dialog", u"Preview", None))
        self.lblLabelPreview.setText("")
        self.btnGetText.setText("")
        self.btnPreview.setText("")
        self.btnCancel.setText("")
        self.btnAccept.setText("")
    # retranslateUi




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This is the depythontex wrapper script.  It automatically detects the version
of Python, and then imports the correct code from depythontex2.py or
depythontex3.py.  It is intended for use with the default Python installation
on your system.  If you wish to use a different version of Python, you could
launch depythontex2.py or depythontex3.py directly.  The version of Python
does not matter for depythontex, since no code is executed.

Copyright (c) 2013-2026, Geoffrey M. Poore
All rights reserved.
Licensed under the BSD 3-Clause License:
    http://www.opensource.org/licenses/BSD-3-Clause

'''

import sys
if sys.version_info[0] == 2:
    import depythontex2 as depythontex
elif sys.version_info[0] == 3:
    import depythontex3 as depythontex



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

'''
PythonTeX depythontex script.

This script takes a LaTeX document that uses the PythonTeX package and
creates a new document that does not depend on PythonTeX.  It substitutes all
externally generated content into a copy of the original LaTeX document.
This is useful when you need a document that relies on few external packages
or custom macros (for example, for submission to a journal or conversion to
another document format).

If you just want to share a document that uses PythonTeX, keep in mind that
the document can be modified and compiled just like a regular LaTeX document,
without needing Python or any other external tools, so long as the following
conditions are met:

  * A copy of pythontex.sty is included with the document.
  * The pythontex-files-<name> directory is included with the document.
  * The PythonTeX-specific parts of the document are not modified.

To work, this script requires that the original LaTeX document be compiled
with the package option `depythontex`.  That creates an auxiliary file with
the extension .depytx that contains information about all content that needs
to be substituted.

This script is purposely written in a simple, largely linear form to
facilitate customization.  Most of the key substitutions are performed by a
few functions defined near the beginning of the script, so if you need custom
substitutions, you should begin there.  By default, all typeset code is
wrapped in `\verb` commands and verbatim environments, since these have the
greatest generality.  However, the command-line option --listing allows code
to be typeset with the fancyvrb, listings, minted, or PythonTeX packages
instead.

The script automatically extracts all arguments of all commands and
environments that it replaces, so that these are available if desired for
customized substitution.  Two additional pieces of information are also
available for any typeset code:  the Pygments lexer (often the same as the
language) and the starting line number (if line numbering was used).

Keep in mind that some manual adjustments may be required after a document is
depythontex'ed.  While depythontex attempts to create an exact copy of the
original document, in many cases an identical copy is impossible.  For
example, typeset code may have a different appearance or layout when it is
typeset with a different package.


Copyright (c) 2013-2026, Geoffrey M. Poore
All rights reserved.
Licensed under the BSD 3-Clause License:
    http://www.opensource.org/licenses/BSD-3-Clause

'''


# Imports
#// Python 2
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
#\\ End Python 2
import sys
import os
#// Python 2
if sys.version_info.major != 2:
    sys.exit('This version of the PythonTeX script requires Python 2.')
#\\ End Python 2
#// Python 3
#if sys.version_info.major != 3:
#    sys.exit('This version of the PythonTeX script requires Python 3.')
#\\ End Python 3

#// Python 2
from io import open
input = raw_input
#\\ End Python 2
import argparse
from collections import defaultdict
from re import match, sub, search
import textwrap
import codecs


# Script parameters
# Version
__version__ = '0.19'


# Functions and parameters for customizing the script output

# Style or package for replacing code listings
# This is actually set via command-line option --listing
# It is created here simply for reference
listing = None  #'verbatim', 'fancyvrb', 'listings', 'minted', 'pythontex'

# List of things to add to the preamble
# It can be appended to via the command-line option --preamble
# It is also appended to based on the code listing style that is used
# And it could be manually edited here as well, as long as it remains a list
preamble_additions = list()

# Lexer dict
# If you are using Pygments lexers that don't directly correspond to the
# languages used by the listings package, you can submit replacements via the
# command line option --lexer-dict, or edit this dict manually here.  When
# listings is used, all lexers are checked against this dict to see if a
# substitution should be made.  This approach could easily be modified to
# work with another, non-Pygments highlighting package.
lexer_dict = dict()


def replace_code_cmd(name, arglist, linenum, code_replacement,
                     code_replacement_mode, after, lexer, firstnumber):
    '''
    Typeset code from a command with a command.

    It is only ever called if there is indeed code to typeset.

    Usually, code from a command is also typeset with a command.  This
    function primarily deals with that case.  In cases where code from a
    command is typeset with an environment (for example, `\inputpygments`),
    this function performs some preprocessing and then uses
    replace_code_env() to do the real work.  This approach prevents the two
    functions from unnecessarily duplicating each other, while still giving
    the desired output.

    Args:
        name (str):  name of the command
        arglist (list, of str/None):  all arguments given to the original
            command; the last argument is what is typeset, unless a
            code_replacement is specified or other instructions are given
        linenum (int):  line number in the original TeX document
        code_replacement (str/None):  replacement for the code; usually None
            for commands, because typically the code to be typeset is the
            last argument passed to the command, rather than something
            captured elsewhere (like the body of an environment) or something
            preprocessed (like a console environment's content)
        code_replacement_mode (str/None):  mode in which the replacement is
            to be typeset; raw/None (as TeX; generally unused for code),
            verb (inline), or verbatim (environment)
        after (str):  text immediately following the command; usually
            shouldn't be needed
        lexer (str/None):  Pygments lexer
    Returns:
        (replacement, after) (tuple, of str)

    '''
    # Get the correct replacement
    if code_replacement is None:
        code_replacement = arglist[-1]

    # We only consider two possible modes of typesetting, verbatim and inline
    # verbatim
    if code_replacement_mode == 'verbatim':
        # Sometimes we must replace a command with an environment, for
        # example, for `\inputpygments`

        # Make sure the introduction of an environment where a command was
        # previously won't produce errors with following content; make sure
        # that any following content is on a separate line
        if bool(match('[ \t]*\S', after)):
            after = '\n' + after
        # Rather than duplicating much of replace_code_env(), just use it
        return replace_code_env(name, arglist, linenum, code_replacement,
                                code_replacement_mode, after, lexer, firstnumber)
    else:
        # Usually, we're replacing a command with a command

        # Wrap the replacement in appropriate delimiters
        if (listing in ('verbatim', 'fancyvrb', 'minted') or
                (listing in ('listings', 'pythontex') and
                ('{' in code_replacement or '}' in code_replacement))):
            for delim in ('|', '/', '`', '!', '&', '#', '@', ':', '%', '~', '$',
                          '=', '+', '-', '^', '_', '?', ';'):
                if delim not in code_replacement:
                    break
            code_replacement = delim + code_replacement + delim
        else:
            code_replacement = '{' + code_replacement + '}'
        # Assemble the actual replacement
        if listing in ('verbatim', 'minted'): # `\mint` isn't for inline use
            code_replacement = r'\verb' + code_replacement
        elif listing == 'fancyvrb':
            code_replacement = r'\Verb' + code_replacement
        elif listing == 'listings':
            if lexer is None:
                code_replacement = r'\lstinline[language={}]' + code_replacement
            else:
                if lexer in lexer_dict:
                    lexer = lexer_dict[lexer]
                code_replacement = r'\lstinline[language=' + lexer + ']' + code_replacement
        elif listing == 'pythontex':
            if lexer is None:
                code_replacement = r'\pygment{text}' + code_replacement
            else:
                code_replacement = r'\pygment{' + lexer + '}' + code_replacement
        return (code_replacement, after)


def replace_code_env(name, arglist, linenum, code_replacement,
                     code_replacement_mode, after, lexer, firstnumber):
    '''
    Typeset code from an environment with an environment.

    It is only ever called if there is indeed code to typeset.

    Usually it is only used to typeset code from an environment.  However,
    some commands bring in code that must be typeset as an environment.  In
    those cases, replace_code_cmd() is called initially, and after it
    performs some preprocessing, this function is called.  This approach
    avoids unnecessary duplication between the two functions.

    Args:
        name (str):  name of the environment
        arglist (list, of str/None):  all arguments given to the original
            environment
        linenum (int):  line number in the original TeX document where
            the environment began
        code_replacement (str):  replacement for the code; unlike the case of
            commands, this is always not None if the function is called
        code_replacement_mode (str/None):  mode in which the replacement is
            to be typeset; raw/None (as TeX; generally unused for code),
            verb (inline), or verbatim (environment)
        after (str):  text immediately following the environment; usually
            shouldn't be needed
        lexer (str/None):  Pygments lexer
        firstnumber (str/None):  the first number of the listing, if the listing
            had numbered lines
    Returns:
        (replacement, after) (tuple, of str)

    '''
    # Currently, there is no need to test for code_replacement_mode, because
    # this function is only ever called if the mode is 'verbatim'.  That may
    # change in the future, but it seems unlikely that code entered in an
    # environment would end up typeset with a command.
    if listing == 'verbatim':
        pre = '\\begin{verbatim}'
        post = '\\end{verbatim}'
    elif listing == 'fancyvrb':
        if firstnumber is None:
            pre = '\\begin{Verbatim}'
        else:
            pre = '\\begin{{Verbatim}}[numbers=left,firstnumber={0}]'.format(firstnumber)
        post = '\\end{Verbatim}'
    elif listing == 'listings':
        if lexer is None:
            if firstnumber is None:
                pre = '\\begin{lstlisting}[language={}]'
            else:
                pre = '\\begin{{lstlisting}}[language={{}},numbers=left,firstnumber={0}]'.format(firstnumber)
        else:
            if lexer in lexer_dict:
                lexer = lexer_dict[lexer]
            if firstnumber is None:
                pre = '\\begin{{lstlisting}}[language={0}]'.format(lexer)
            else:
                pre = '\\begin{{lstlisting}}[language={0},numbers=left,firstnumber={1}]'.format(lexer, firstnumber)
        post = '\\end{lstlisting}'
    elif listing == 'minted':
        if lexer is None:
            if firstnumber is None:
                pre = '\\begin{minted}{text}'
            else:
                pre = '\\begin{{minted}}[linenos,firstnumber={0}]{{text}}'.format(firstnumber)
        else:
            if firstnumber is None:
                pre = '\\begin{{minted}}{{{0}}}'.format(lexer)
            else:
                pre = '\\begin{{minted}}[linenos,firstnumber={0}]{{{1}}}'.format(firstnumber, lexer)
        post = '\\end{minted}'
    elif listing == 'pythontex':
        if lexer is None:
            if firstnumber is None:
                pre = '\\begin{pygments}{text}'
            else:
                pre = '\\begin{{pygments}}[numbers=left,firstnumber={0}]{{text}}'.format(firstnumber)
        else:
            if firstnumber is None:
                pre = '\\begin{{pygments}}{{{0}}}'.format(lexer)
            else:
                pre = '\\begin{{pygments}}[numbers=left,firstnumber={0}]{{{1}}}'.format(firstnumber, lexer)
        post = '\\end{pygments}'
    code_replacement = pre + code_replacement + post
    return (code_replacement, after)


# We will need to issue a warning every time that a substitution of printed
# content results in a forced double space.  We could just do this as we go,
# but it's easier for the user to read if we just collect all the warnings
# of this type, and print them once.
forced_double_space_list = list()


def replace_print_cmd(name, arglist, linenum,
                      print_replacement, print_replacement_mode, source,
                      after):
    '''
    Typeset printed content from a command.

    It is only ever called if there is indeed printed content to typeset.

    Args:
        name (str):  name of the command
        arglist (list, of str/None):  all arguments given to the original
            command
        linenum (int):  line number in the original TeX document
        print_replacement (str):  printed content, read directly from file
            into a single string
        print_replacement_mode (str/None):  mode in which the replacement is
            to be typeset; raw/None (as TeX), inlineverb (or v) (as inline),
            or verb (as environment)
        source (str/None):  source of the replacement content
        after (str):  text immediately following the command; important in
            some situations, because spacing can depend on what's next
    Returns:
        (replacement, after) (tuple, of str)

    '''
    if print_replacement_mode == 'verb':
        if print_replacement.count('\n') > 1:
            print('* DePythonTeX error:')
            print('    Attempt to print multiple lines of content near line ' + str(linenum))
            print('    This is not possible in inline verbatim mode')
            sys.exit(1)
        print_replacement = print_replacement.rstrip('\n')
        for delim in ('|', '/', '`', '!', '&', '#', '@', ':', '%', '~', '$',
                      '=', '+', '-', '^', '_', '?', ';'):
            if delim not in print_replacement:
                break
        print_replacement = r'\verb' + delim + print_replacement + delim
    elif print_replacement_mode == 'verbatim':
        if bool(match('\s*?\n', after)):
            # Usually, we would end the verbatim environment with a newline.
            # This is fine if there is content in `after` before the next
            # newline---in fact, it's desirable, because the verbatim package
            # doesn't allow for content on the same line as the end of the
            # environment.  But if `after` is an empty line, then adding a
            # newline will throw off spacing and must be avoided
            print_replacement = '\\begin{verbatim}\n' + print_replacement + '\\end{verbatim}'
        else:
            print_replacement = '\\begin{verbatim}\n' + print_replacement + '\\end{verbatim}\n'
    else:
        # When printed content from a file is included as LaTeX code, we have
        # to be particularly careful to ensure that the content produces the
        # same output when substituted as when brought in by `\input`.  In
        # particular, `\input` strips newlines from each line of content and
        # adds a space at the end of each line.  This space is inside the
        # `\input`, so it will not merge with following spaces.  So when we
        # substitute the content, sometimes we need to replace the final
        # newline with a space that cannot be gobbled.
        #
        # It gets more complicated.  This final space is often not
        # desirable.  It can be prevented by either printing an `\endinput`
        # command, to terminate the `\input`, or printing a percent
        # character % in the last line of the content, which comments out the
        # final newline.  So we must check for `\endinput` anywhere in
        # printed content, and % in the final line, and remove any content
        # after them.  It's also possible that the print is followed by
        # an `\unskip` that eats the space, so we need to check for that too.
        #
        # It turns out that the same approach is needed when a command like
        # `\py` brings in content ending in a newline
        if (print_replacement.endswith('\\endinput\n') and
                not print_replacement.endswith('\\string\\endinput\n')):
            # If `\endinput` is present, everything from it on should be
            # discarded, unless the `\endinput` is not actually a command
            # but rather a typeset name (for example, `\string\endinput` or
            # `\verb|\endinput|`).  It's impossible to check for all cases in
            # which `\endinput` is not a command (at least, without actually
            # using LaTeX), and even checking for most of them would require
            # a good bit of parsing.  We assume that `\endinput`, as a
            # command, will only ever occur at the immediate end of the
            # printed content.  Later, we issue a warning in case it appears
            # anywhere else.
            print_replacement = print_replacement.rsplit(r'\endinput', 1)[0]
        elif (print_replacement.endswith('%\n') and
                not print_replacement.endswith('\\%\n') and
                not print_replacement.endswith('\\string%\n')):
            # Perform an analogous check for a terminating percent characer %.
            # This case would be a bit easier to parse fully, since a percent
            # that comments out the last newline would have to be in the
            # final line of the replacement.  But it would still be
            # very difficult to perform a complete check.  Later, we issue a
            # warning if there is reason to think that a percent character
            # was active in the last line.
            print_replacement = print_replacement.rsplit(r'%', 1)[0]
        elif print_replacement.endswith('\n'):
            # We can't just use `else` because that would catch content
            # from `\py` and similar
            # By default, LaTeX strips newlines and adds a space at the end
            # of each line of content that is brought in by `\input`.  This
            # may or may not be desirable, but we replicate the effect here
            # for consistency with the original document.  We use `\space{}`
            # because plain `\space` would gobble a following space, which
            # isn't consistent with the `\input` behavior being replicated.
            if bool(match(r'\\unskip\s+\S', after)):
                # If there's an `\unskip`, fix the spacing and remove the
                # `\unskip`.  Since this is inline, the `\unskip` must
                # immediately follow the command to do any good; otherwise,
                # it eliminates spaces that precede it, but doesn't get into
                # the `\input` content.
                print_replacement = print_replacement.rstrip(' \t\n')
                after = sub(r'^\\unskip\s+', '', after)
            elif bool(match('\S', after)):
                # If the next character is not whitespace, we can just leave
                # the `\n`, and it will yield a space.
                pass
            elif bool(match('\s*$', after)):
                # If the rest of the current line, and the next line, are
                # whitespace, we will get the correct spacing without needing
                # `\space{}`.  We could leave `\n`, but it would be
                # extraneous whitespace.
                print_replacement = print_replacement[:-1]
            else:
                # Otherwise, we do need to insert `\space{}`
                # We keep the newline at the end of printed content, in case
                # it's at the end of an environment, and thus is needed to
                # protect the following content
                print_replacement += '\\space{}'
                after = sub('^\s+', '', after)
                forced_double_space_list.append((name, linenum))
        else:
            if bool(match('\s+\S', after)):
                # If the following line starts with whitespace, replace it
                # with a newline, to protect in the event that the printed
                # content ended with an end-of-environment command
                after = sub('^\s+', '\n', after)
        # Issue warnings, if warranted
        # Warn about `\endinput`
        if (r'\endinput' in print_replacement and
                print_replacement.count(r'\endinput') != print_replacement.count(r'\string\endinput')):
            print('* DePythonTeX warning:')
            print('    "\\endinput" was present in printed content near line ' + str(linenum))
            print('    If this "\\endinput" was verbatim, you have nothing to worry about')
            print('    If this "\\endinput" is to be active, it should be printed last')
            print('    If you need "\\endinput" elsewhere, customize depythontex.py')
        # Warn if it looks like there are active `%` that could comment
        # out part of the original document.  We only need to check the
        # last line of printed content, because only there could
        # percent characters escape from their original confines within
        # `\input`, and comment out part of the document.
        if print_replacement.endswith('\n'):
            if print_replacement.count('\n') > 1:
                last_line = print_replacement.rsplit('\n', 2)[1]
            else:
                last_line = print_replacement
            if last_line.count('%') != last_line.count(r'\%') + last_line.count(r'\string%'):
                print('* DePythonTeX warning:')
                print('    One or more percent characters are on the last line of ')
                print('    printed content near line ' + str(linenum))
                print('    If they are not verbatim, they could comment out the following newline')
        else:
            if '\n' in print_replacement:
                last_line = print_replacement.rsplit('\n', 1)[1]
            else:
                last_line = print_replacement
            if last_line.count('%') != last_line.count(r'\%') + last_line.count(r'\string%'):
                print('* DePythonTeX warning:')
                print('    One or more percent characters are on the last line of ')
                print('    printed content near line ' + str(linenum))
                print('    If they are not verbatim, they could comment out the following text')
        # Warn if there were `\unskip`'s in the output, in case they weren't
        # positioned correctly
        if bool(match(r'\s*\\unskip', after)):
            print('* DePythonTeX warning:')
            print('    "\\unskip" survived substitution near line ' + str(linenum))
            print('    If it should have adjusted the spacing of printed content')
            print('    you should double-check the spacing')
    return (print_replacement, after)


def replace_print_env(name, arglist, linenum,
                      print_replacement, print_replacement_mode, source,
                      after):
    '''
    Typeset printed content from an environment.

    It is only ever called if there is indeed printed content to typeset.

    This should be similar to replace_print_cmd().  The main difference is
    that the environment context typically ends with a newline, so
    substitution has to be a little different to ensure that spacing after
    the environment isn't modified.

    Args:
        name (str):  name of the environment
        arglist (list, of str/None):  all arguments given to the original
            environment
        linenum (int):  line number in the original TeX document where the
            environment began
        print_replacement (str):  printed content, read directly from file
            into a single string
        print_replacement_mode (str/None):  mode in which the replacement is
            to be typeset; raw/None (as TeX), inlineverb (or v) (as inline),
            or verb (as environment)
        source (str/None):  source of the replacement content
        after (str):  text immediately following the command; important in
            some situations, because spacing can depend on what's next
    Returns:
        (replacement, after) (tuple, of str)

    #### The inlineverb and verb modes should work, but haven't been tested
    since there are currently no environments that use them; they are only
    used by `\printpythontex`, which is a command.
    '''
    if print_replacement_mode == 'verb':
        if print_replacement.count('\n') > 1:
            print('* DePythonTeX error:')
            print('    Attempt to print multiple lines of content near line ' + str(linenum))
            print('    This is not possible in inline verbatim mode')
            sys.exit(1)
        print_replacement = print_replacement.rstrip('\n')
        for delim in ('|', '/', '`', '!', '&', '#', '@', ':', '%', '~', '$',
                      '=', '+', '-', '^', '_', '?', ';'):
            if delim not in print_replacement:
                break
        print_replacement = r'\verb' + delim + print_replacement + delim
        if not bool(match('[ \t]+\S', after)):
            # If there is text on the same line as the end of the
            # environment, we're fine (this is unusual).  Otherwise,
            # we need to toss the newline at the end of the environment
            # and gobble leading spaces.  Leading spaces need to be
            # gobbled because previously they were at the beginning of a
            # line, where they would have been discarded.
            if not bool(match('\s*$', after)):
                after = sub('^\s*?\n\s*', '', after)
    elif print_replacement_mode == 'verbatim':
        if bool(match('\s*?\n', after)):
            # Usually, we would end the verbatim environment with a newline.
            # This is fine if there is content in `after` before the next
            # newline---in fact, it's desirable, because the verbatim package
            # doesn't allow for content on the same line as the end of the
            # environment.  But if `after` is an empty line, then adding a
            # newline will throw off spacing and must be avoided
            print_replacement = '\\begin{verbatim}\n' + print_replacement + '\\end{verbatim}'
        else:
            print_replacement = '\\begin{verbatim}\n' + print_replacement + '\\end{verbatim}\n'
    else:
        # When printed content is included as LaTeX code, we have to be
        # particularly careful to ensure that the content produces the same
        # output when substituted as when brought in by `\input`.  In
        # particular, `\input` strips newlines from each line of content and
        # adds a space at the end of each line.  This space is inside the
        # `\input`, so it will not merge with following spaces.  So when we
        # substitute the content, sometimes we need to replace the final
        # newline with a space that cannot be gobbled.
        #
        # It gets more complicated.  This final space is often not
        # desirable.  It can be prevented by either printing an `\endinput`
        # command, to terminate the `\input`, or printing a percent
        # character % in the last line of the content, which comments out the
        # final newline.  So we must check for `\endinput` anywhere in
        # printed content, and % in the final line, and remove any content
        # after them.  It's also possible that the print is followed by
        # an `\unskip` that eats the space, so we need to check for that too.
        if (print_replacement.endswith('\\endinput\n') and
                not print_replacement.endswith('\\string\\endinput\n')):
            # If `\endinput` is present, everything from it on should be
            # discarded, unless the `\endinput` is not actually a command
            # but rather a typeset name (for example, `\string\endinput` or
            # `\verb|\endinput|`).  It's impossible to check for all cases in
            # which `\endinput` is not a command (at least, without actually
            # using LaTeX), and even checking for most of them would require
            # a good bit of parsing.  We assume that `\endinput`, as a
            # command, will only ever occur at the immediate end of the
            # printed content.  Later, we issue a warning in case it appears
            # anywhere else.
            print_replacement = print_replacement.rsplit(r'\endinput', 1)[0]
            if not bool(match('[ \t]+\S', after)):
                # If there is text on the same line as the end of the
                # environment, we're fine (this is unusual).  Otherwise,
                # we need to toss the newline at the end of the environment
                # and gobble leading spaces.  Leading spaces need to be
                # gobbled because previously they were at the beginning of a
                # line, where they would have been discarded.
                if not bool(match('\s*$', after)):
                    after = sub('^\s*?\n\s*', '', after)
        elif (print_replacement.endswith('%\n') and
                not print_replacement.endswith('\\%\n') and
                not print_replacement.endswith('\\string%\n')):
            # Perform an analogous check for a terminating percent characer %.
            # This case would be a bit easier to parse fully, since a percent
            # that comments out the last newline would have to be in the
            # final line of the replacement.  But it would still be
            # very difficult to perform a complete check.  Later, we issue a
            # warning if there is reason to think that a percent character
            # was active in the last line.
            print_replacement = print_replacement.rsplit(r'%', 1)[0]
            if not bool(match('[ \t]+\S', after)):
                # If there is text on the same line as the end of the
                # environment, we're fine (this is unusual).  Otherwise,
                # we need to toss the newline at the end of the environment
                # and gobble leading spaces.  Leading spaces need to be
                # gobbled because previously they were at the beginning of a
                # line, where they would have been discarded.
                if not bool(match('\s*$', after)):
                    after = sub('^\s*?\n\s*', '', after)
        else:
            # By default, LaTeX strips newlines and adds a space at the end
            # of each line of content that is brought in by `\input`.  This
            # may or may not be desirable, but we replicate the effect here
            # for consistency with the original document.  We use `\space{}`
            # because plain `\space` would gobble a following space, which
            # isn't consistent with the `\input` behavior being replicated.
            if bool(match(r'\s*\\unskip\s+\S', after)):
                # If there's an `\unskip`, fix the spacing and remove the
                # `\unskip`
                print_replacement = print_replacement.rstrip(' \t\n')
                after = sub(r'^\s*\\unskip\s+', '', after)
            elif bool(match('[ \t]+\S', after)):
                # If the next character after the end of the environment is
                # not whitespace (usually not allowed), we can just leave
                # the `\n` in printed content, and it will yield a space.
                # So we need do nothing.  But if there is text on that line
                # we need `\space{}`.
                after = sub('^\s+', '\\space', after)
                forced_double_space_list.append((name, linenum))
            else:
                # If the line at the end of the environment is blank,
                # we can just discard it and keep the newline at the end of
                # the printed content; the newline gives us the needed space
                after = after.split('\n', 1)[1]
        # Issue warnings, if warranted
        # Warn about `\endinput`
        if (r'\endinput' in print_replacement and
                print_replacement.count(r'\endinput') != print_replacement.count(r'\string\endinput')):
            print('* DePythonTeX warning:')
            print('    "\\endinput" was present in printed content near line ' + str(linenum))
            print('    If this "\\endinput" was verbatim, you have nothing to worry about')
            print('    If this "\\endinput" is to be active, it should be printed last')
            print('    If you need "\\endinput" elsewhere, customize depythontex.py')
        # Warn if it looks like there are active `%` that could comment
        # out part of the original document.  We only need to check the
        # last line of printed content, because only there could
        # percent characters escape from their original confines within
        # `\input`, and comment out part of the document.
        if print_replacement.endswith('\n'):
            if print_replacement.count('\n') > 1:
                last_line = print_replacement.rsplit('\n', 2)[1]
            else:
                last_line = print_replacement
            if last_line.count('%') != last_line.count(r'\%') + last_line.count(r'\string%'):
                print('* DePythonTeX warning:')
                print('    One or more percent characters are on the last line of ')
                print('    printed content near line ' + str(linenum))
                print('    If they are not verbatim, they could comment out the following newline')
        else:
            if '\n' in print_replacement:
                last_line = print_replacement.rsplit('\n', 1)[1]
            else:
                last_line = print_replacement
            if last_line.count('%') != last_line.count(r'\%') + last_line.count(r'\string%'):
                print('* DePythonTeX warning:')
                print('    One or more percent characters are on the last line of ')
                print('    printed content near line ' + str(linenum))
                print('    If they are not verbatim, they could comment out the following text')
        # Warn if there were `\unskip`'s in the output, in case they weren't
        # positioned correctly
        if bool(match(r'\s*\\unskip', after)):
            print('* DePythonTeX warning:')
            print('    "\\unskip" survived substitution near line ' + str(linenum))
            print('    If it should have adjusted the spacing of printed content')
            print('    you should double-check the spacing')
    return (print_replacement, after)




# Deal with argv
# Parse argv
parser = argparse.ArgumentParser()
parser.add_argument('--version', action='version',
                    version='DePythonTeX {0}'.format(__version__))
parser.add_argument('--encoding', default='utf-8',
                    help='encoding for all text files (see codecs module for encodings)')
parser.add_argument('--overwrite', default=False, action='store_true',
                    help='overwrite existing output, if it exists (off by default)')
parser.add_argument('--listing', default='verbatim',
                    choices=('verbatim', 'fancyvrb', 'listings', 'minted', 'pythontex'),
                    help='style or package used for typesetting code')
parser.add_argument('--lexer-dict', default=None,
                    help='add mappings from Pygments lexer names to the language names of other highlighting packages; should be a comma-separated list of the form "<Pygments lexer>:<language>, <Pygments lexer>:<language>, ..."')
parser.add_argument('--preamble', default=None,
                    help='line of commands to add to output preamble')
parser.add_argument('--graphicspath', default=False, action='store_true',
                    help=r'Add the outputdir to the graphics path, by modifying an existing \graphicspath command or adding one.')
parser.add_argument('-o', '--output', default=None,
                    help='output file')
parser.add_argument('TEXNAME',
                    help='LaTeX file')
args = parser.parse_args()

# Process argv
encoding = args.encoding
listing = args.listing
if args.preamble is not None:
    preamble_additions.append(args.preamble)
if args.lexer_dict is not None:
    args.lexer_dict = args.lexer_dict.replace(' ', '').replace("'", "").replace('"','').strip('{}')
    for entry in args.lexer_dict.split(','):
        k, v = entry.split(':')
        lexer_dict[k] = v
if args.listing == 'verbatim':
    # In some contexts, the verbatim package might be desirable.
    # But we assume that the user wants minimal packages.
    # Also, the default verbatim environment doesn't allow text to follow the
    # end-of-environment command.
    # If the verbatim package is ever desired, simply uncomment the following:
    # preamble_additions.append('\\usepackage{verbatim}')
    pass
elif args.listing == 'fancyvrb':
    preamble_additions.append('\\usepackage{fancyvrb}')
elif args.listing == 'listings':
    preamble_additions.append('\\usepackage{listings}')
elif args.listing == 'minted':
    preamble_additions.append('\\usepackage{minted}')
elif args.listing == 'pythontex':
    preamble_additions.append('\\usepackage{pythontex}')




# Let the user know things have started
if args.output is not None:
    print('This is DePythonTeX {0}'.format(__version__))
    sys.stdout.flush()




# Make sure we have a valid texfile
texfile_name = os.path.expanduser(os.path.normcase(args.TEXNAME))
if not os.path.isfile(texfile_name):
    resolved = False
    if not texfile_name.endswith('.tex'):
        for ext in ('.tex', '.ltx', '.dtx'):
            if os.path.isfile(texfile_name + ext):
                texfile_name = texfile_name + ext
                resolved = True
                break
    if not resolved:
        print('* DePythonTeX error:')
        print('    Could not locate file "' + texfile_name + '"')
        sys.exit(1)
# Make sure we have a valid outfile
if args.output is not None:
    outfile_name = os.path.expanduser(os.path.normcase(args.output))
    if not args.overwrite and os.path.isfile(outfile_name):
        print('* DePythonTeX warning:')
        print('    Output file "' + outfile_name + '" already exists')
        ans = input('    Do you want to overwrite this file? [y,n]\n    ')
        if ans != 'y':
            sys.exit(1)
# Make sure the .depytx file exists
depytxfile_name = texfile_name.rsplit('.')[0] + '.depytx'
if not os.path.isfile(depytxfile_name):
    print('* DePythonTeX error:')
    print('    Could not find DePythonTeX auxiliary file "' + depytxfile_name + '"')
    print('    Use package option depythontex to create it')
    sys.exit(1)




# Start opening files and loading data
# Read in the LaTeX file
# We read into a list with an empty first entry, so that we don't have to
# worry about zero indexing when comparing list index to file line number
f = open(texfile_name, 'r', encoding=encoding)
tex = ['']
tex.extend(f.readlines())
f.close()
# Load the .depytx
f = open(depytxfile_name, 'r', encoding=encoding)
depytx = f.readlines()
f.close()
# Process the .depytx by getting the settings contained in the last few lines
settings = dict()
n = len(depytx) - 1
while depytx[n].startswith('=>DEPYTHONTEX:SETTINGS#'):
    content = depytx[n].split('#', 1)[1].rsplit('#', 1)[0]
    k, v = content.split('=', 1)
    if v in ('true', 'True'):
        v = True
    elif v in ('false', 'False'):
        v = False
    settings[k] = v
    depytx[n] = ''
    n -= 1
# Check .depytx version to make sure it is compatible
if settings['version'] != __version__:
    print('* DePythonTeX warning:')
    print('    Version mismatch with DePythonTeX auxiliary file')
    print('    Do a complete compile cycle to update the auxiliary file')
    print('    Attempting to proceed')
# Go ahead and open the outfile, even though we don't need it until the end
# This lets us change working directories for convenience without worrying
# about having to modify the outfile path
if args.output is not None:
    outfile = open(outfile_name, 'w', encoding=encoding)




# Change working directory to the document directory
# Technically, we could get by without this, but that would require a lot of
# path modification.  This way, we can just use all paths straight out of the
# .depytx without any modification, which is much simpler and less error-prone.
if os.path.split(texfile_name)[0] != '':
    os.chdir(os.path.split(texfile_name)[0])




# Open and process the file of macros
# Read in the macros
if os.path.isfile(os.path.expanduser(os.path.normcase(settings['macrofile']))):
    f = open(os.path.expanduser(os.path.normcase(settings['macrofile'])), 'r', encoding=encoding)
    macros = f.readlines()
    f.close()
else:
    print('* DePythonTeX error:')
    print('    The macro file could not be found:')
    print('      "' + settings['macrofile'] + '"')
    print('    Run PythonTeX to create it')
    sys.exit(1)
# Create a dict for storing macros
macrodict = defaultdict(list)
# Create variables for keeping track of whether we're inside a macro or
# environment
# These must exist before we begin processing
inside_macro = False
inside_environment = False
# Loop through the macros, and extract everything
# We just extract content; we get content wrappers later, when we process all
# substituted content
for line in macros:
    if inside_macro:
        # If we're in a macro, look for the end-of-macro command
        if r'\endpytx@SVMCR' in line:
            # If the current line contains the end-of-macro command, split
            # off any content that comes before it.  Also reset
            # `inside_macro`.
            macrodict[current_macro].append(line.rsplit(r'\endpytx@SVMCR', 1)[0])
            inside_macro = False
        else:
            # If the current line doesn't end the macro, we add the whole
            # line to the macro dict
            macrodict[current_macro].append(line)
    elif inside_environment:
        if line.startswith(end_environment):
            # If the environment is ending, we reset inside_environment
            inside_environment = False
        else:
            # If we're still in the environment, add the current line to the
            # macro dict
            macrodict[current_macro].append(line)
    else:
        # If we're not in a macro or environment, we need to figure out which
        # we are dealing with (if either; there are blank lines in the macro
        # file to increase readability).  Once we've determined which one,
        # we need to get its name and extract any content.
        if line.startswith(r'\begin{'):
            # Any \begin will indicate a use of fancyvrb to save verbatim
            # content, since that is the only time an environment is used in
            # the macro file.  All other content is saved in a standard macro.
            # We extract the name of the macro in which the verbatim content
            # is saved.
            current_macro = line.rsplit('{', 1)[1].rstrip('}\n')
            inside_environment = True
            # We assemble the end-of-environment string we will need to look
            # for.  We don't assume any particular name, for generality.
            end_environment = r'\end{' + line.split('}', 1)[0].split('{', 1)[1] + '}'
            # Code typset in an environment needs to have a leading newline,
            # because the content of a normal verbatim environment keeps its
            # leading newline.
            macrodict[current_macro].append('\n')
        elif line.startswith(r'\pytx@SVMCR{'):
            # Any regular macro will use `\pytx@SVMCR`
            current_macro = line.split('{', 1)[1].split('}', 1)[0]
            inside_macro = True
            # Any content will always be on the next line, so we don't need
            # to check for it




# Do the actual processing
# Create a variable for keeping track of the current line in the LaTeX file
# Start at 1, since the first entry in the tex list is `''`
texlinenum = 1
# Create a variable for storing the current line(s) we are processing.
# This contains all lines from immediately after the last successfully
# processed line up to and including texlinenum.  We may have to process
# multiple lines at once if a macro is split over multiple lines, etc.
texcontent = tex[texlinenum]
# Create a list for storing processed content.
texout = list()
# Loop through the depytx and process
for n, depytxline in enumerate(depytx):
    if depytxline.startswith('=>DEPYTHONTEX#'):
        # Process info
        depytxcontent = depytxline.split('#', 1)[1].rstrip('#\n')
        depy_type, depy_name, depy_args, depy_typeset, depy_linenum, depy_lexer = depytxcontent.split(':')
        if depy_lexer == '':
            depy_lexer = None

        # Do a quick check on validity of info
        # #### Eventually add 'cp' and 'pc'
        if not (depy_type in ('cmd', 'env') and
                all([letter in ('o', 'm', 'v', 'n', '|') for letter in depy_args]) and
                ('|' not in depy_args or (depy_args.count('|') == 1 and depy_args.endswith('|'))) and
                depy_typeset in ('c', 'p', 'n')):
            print('* PythonTeX error:')
            print('    Invalid \\Depythontex string for operation on line ' + str(depy_linenum))
            print('    The offending string was ' + depytxcontent)
            sys.exit(1)
        # If depy_args contains a `|` to indicate `\obeylines`, strip it and
        # store in a variable.  Create a bool to keep track of obeylines
        # status, which governs whether we can look on the next line for
        # arguments.  (If obeylines is active, a newline terminates the
        # argument search.)
        if depy_args.endswith('|'):
            obeylines = True
            depy_args = depy_args.rstrip('|')
        else:
            obeylines = False
        # Get the line number as an integer
        # We don't have to adjust for zero indexing in tex
        depy_linenum = int(depy_linenum)


        # Check for information passed from LaTeX
        # This will be extra listings information, or replacements to plug in
        code_replacement = None
        code_replacement_mode = None
        print_replacement = None
        print_replacement_mode = None
        firstnumber = None
        source = None
        scan_ahead_line = n + 1
        nextdepytxline = depytx[scan_ahead_line]
        while not nextdepytxline.startswith('=>DEPYTHONTEX#'):
            if nextdepytxline.startswith('LISTING:'):
                listingcontent = nextdepytxline.split(':', 1)[1].rstrip('\n')
                if bool(match(r'firstnumber=\d+$', listingcontent)):
                    firstnumber = listingcontent.split('=', 1)[1]
                else:
                    print('* DePythonTeX error:')
                    print('    Unknown information in listings data on line ' + str(depy_linenum))
                    print('    The listings content was "' + listingcontent + '"')
                    sys.exit(1)
            elif nextdepytxline.startswith('MACRO:'):
                source = 'macro'
                try:
                    typeset, macro = nextdepytxline.rstrip('\n').split(':', 2)[1:]
                except:
                    print('* DePythonTeX error:')
                    print('    Improperly formatted macro information on line ' + str(depy_linenum))
                    print('    The macro information was "' + nextdepytxline + '"')
                    sys.exit(1)
                if macro not in macrodict:
                    print('* DePythonTeX error:')
                    print('    Could not find replacement content for macro "' + macro + '"')
                    print('    This is probably because the document needs to be recompiled')
                    sys.exit(1)
                if typeset == 'c':
                    if depy_type == 'cmd':
                        code_replacement = ''.join(macrodict[macro]).strip('\n')
                    else:
                        code_replacement = ''.join(macrodict[macro])
                elif typeset == 'p':
                    print_replacement = ''.join(macrodict[macro])
                else:
                    print('* DePythonTeX error:')
                    print('    Improper typesetting information for macro information on line ' + str(depy_linenum))
                    print('    The macro information was "' + nextdepytxline + '"')
                    sys.exit(1)
            elif nextdepytxline.startswith('FILE:'):
                source = 'file'
                try:
                    typeset, f_name = nextdepytxline.rstrip('\n').split(':', 2)[1:]
                except:
                    print('* DePythonTeX error:')
                    print('    Improperly formatted file information on line ' + str(depy_linenum))
                    print('    The file information was "' + nextdepytxline + '"')
                    sys.exit(1)
                # Files that are brought in have an optional mode that
                # determines if they need special handling (for example, verbatim)
                if ':mode=' in f_name:
                    f_name, mode = f_name.split(':mode=')
                else:
                    mode = None
                f = open(os.path.expanduser(os.path.normcase(f_name)), 'r', encoding=encoding)
                replacement = f.read()
                f.close()
                if typeset == 'c':
                    code_replacement_mode = mode
                    if depy_type == 'cmd' and code_replacement_mode != 'verbatim':
                        # Usually, code from commands is typeset with commands
                        # and code from environments is typeset in
                        # environments.  The except is code from commands
                        # that bring in external files, like `\inputpygments`
                        code_replacement = replacement
                    else:
                        # If we're replacing an environment of code with a
                        # file, then we lose the newline at the beginning
                        # of the environment, and need to get it back.
                        code_replacement = '\n' + replacement
                elif typeset == 'p':
                    print_replacement_mode = mode
                    print_replacement = replacement
                else:
                    print('* DePythonTeX error:')
                    print('    Improper typesetting information for file information on line ' + str(depy_linenum))
                    print('    The file information was "' + nextdepytxline + '"')
                    sys.exit(1)
            # Increment the line in depytx to check for more information
            # from LaTeX
            scan_ahead_line += 1
            if scan_ahead_line == len(depytx):
                break
            else:
                nextdepytxline = depytx[scan_ahead_line]


        # If the line we're looking for is within the range currently held by
        # texcontent, do nothing.  Otherwise, transfer content from tex
        # to texout until we get to the line of tex that we're looking for
        if depy_linenum > texlinenum:
            texout.append(texcontent)
            texlinenum += 1
            while texlinenum < depy_linenum:
                texout.append(tex[texlinenum])
                texlinenum += 1
            texcontent = tex[texlinenum]


        # Deal with arguments
        # All arguments are parsed and stored in a list variables, even if
        # they are not used, for completeness; this makes it easy to add
        # functionality
        # Start by splitting the current line into what comes before the
        # command or environment, and what is after it
        if depy_type == 'cmd':
            try:
                before, after = texcontent.split('\\' + depy_name, 1)
            except:
                print('* DePythonTeX error:')
                print('    Could not find command "' + depy_name + '" on line ' + str(depy_linenum))
                sys.exit(1)
        else:  # depy_type == 'env':
            try:
                before, after = texcontent.split(r'\begin{' + depy_name + '}', 1)
            except:
                print('* DePythonTeX error:')
                print('    Could not find environment "' + depy_name + '" on line ' + str(depy_linenum))
                sys.exit(1)
        # We won't need the content from before the command or environment
        # again, so we go ahead and store it
        texout.append(before)

        # Parse the arguments
        # Create a list for storing the recovered arguments
        arglist = list()
        for argindex, arg in enumerate(depy_args):
            if arg == 'n':
                pass
            elif arg == 'o':
                if after[0] == '[':
                    # Account for possible line breaks before end of arg
                    while ']' not in after:
                        texlinenum += 1
                        after += tex[texlinenum]
                    optarg, after = after[1:].split(']', 1)
                else:
                    if obeylines:
                        # Take into account possible whitespace before arg
                        if bool(match('[ \t]*\[', after)):
                            after = after.split('[', 1)[1]
                            while ']' not in after:
                                texlinenum += 1
                                after += tex[texlinenum]
                            optarg, after = after.split(']', 1)
                        else:
                            optarg = None
                            # If this is the last arg, and it wasn't found,
                            # the macro should eat all whitespace following it
                            if argindex == len(depy_args) - 1:
                                after = sub('^[ \t]*', '', after)
                    else:
                        # Allow peeking ahead a line for the argument
                        if bool(match('\s*$', after)) and after.count('\n') < 2:
                            texlinenum += 1
                            after += tex[texlinenum]
                        # Take into account possible whitespace before arg
                        if bool(match('\s*\[', after)):
                            after = after.split('[', 1)[1]
                            while ']' not in after:
                                texlinenum += 1
                                after += tex[texlinenum]
                            optarg, after = after.split(']', 1)
                        else:
                            optarg = None
                            # Account for eating whitespace afterward, if arg not found
                            if argindex == len(depy_args) - 1:
                                if bool(match('\s*$', after)) and after.count('\n') < 2:
                                    texlinenum += 1
                                    after += tex[texlinenum]
                                if not bool(match('\s*$', after)):
                                    after = sub('^\s*', '', after)
                arglist.append(optarg)
            elif arg == 'm':
                # Account for possible line breaks or spaces before arg
                if after[0] == '{':
                    after = after[1:]
                else:
                    if obeylines:
                        # Account for possible leading whitespace
                        if bool(match('[ \t\f\v]*\{', after)):
                            after = after.split('{', 1)[1]
                        else:
                            print('* DePythonTeX error:')
                            print('    Flawed mandatory argument for "' + depy_name + '" on line ' + str(depy_linenum))
                            sys.exit(1)
                    else:
                        # Peek ahead a line if needed
                        if bool(match('\s*$', after)) and after.count('\n') < 2:
                            texlinenum += 1
                            after += tex[texlinenum]
                        if bool(match('\s*\{', after)):
                            after = after.split('{', 1)[1]
                        else:
                            print('* DePythonTeX error:')
                            print('    Flawed mandatory argument for "' + depy_name + '" on line ' + str(depy_linenum))
                            sys.exit(1)
                # Go through the argument character by character to find the
                # closing brace.
                # If possible, use a very simple approach
                if (r'\{' not in after and r'\}' not in after and
                        r'\string' not in after and
                        after.count('{') + 1 == after.count('}')):
                    pos = 0
                    lbraces = 1
                    rbraces = 0
                    while True:
                        if after[pos] == '{':
                            lbraces += 1
                        elif after[pos] == '}':
                            rbraces += 1
                        if lbraces == rbraces:
                            break
                        pos += 1
                        if pos == len(after):
                            texlinenum += 1
                            after += tex[texlinenum]
                # If a simple parsing approach won't work, parse in much
                # greater depth
                else:
                    pos = 0
                    lbraces = 1
                    rbraces = 0
                    while True:
                        if after[pos] == '{':
                            # If the current character is a brace, we count it
                            lbraces += 1
                            if lbraces == rbraces:
                                break
                            pos += 1
                        elif after[pos] == '}':
                            # If the current character is a brace, we count it
                            rbraces += 1
                            if lbraces == rbraces:
                                break
                            pos += 1
                        elif after[pos:].startswith(r'\string'):
                            # If the current position marks the beginning of `\string`, we
                            # resolve the `\string` command
                            # First, jump ahead to after `\string`
                            pos += 7 #+= len(r'\string')
                            # See if `\string` is followed by a regular macro
                            # If so, jump past it; otherwise, figure out if a
                            # single-character macro, or just a single character, is next,
                            # and jump past it
                            standard_macro = match(r'\\[a-zA-Z]+', line[pos:])
                            if bool(standard_macro):
                                pos += standard_macro.end()
                            elif line[pos] == '\\':
                                pos += 2
                            else:
                                pos += 1
                        elif line[pos] == '\\':
                            # If the current position is a backslash, figure out what
                            # macro is used, and jump past it
                            # The macro must either be a standard alphabetic macro,
                            # or a single-character macro
                            standard_macro = match(r'\\[a-zA-Z]+', line[pos:])
                            if bool(standard_macro):
                                pos += standard_macro.end()
                            else:
                                pos += 2
                        else:
                            pos += 1
                        if pos == len(after):
                            texlinenum += 1
                            after += tex[texlinenum]
                mainarg = after[:pos]
                after = after[pos+1:]
                arglist.append(mainarg)
            elif arg == 'v':
                if after[0] == '{':
                    # Account for the possibility of matched brace delimiters
                    # Not all verbatim commands allow for these
                    pos = 1
                    lbraces = 1
                    rbraces = 0
                    while True:
                        if after[pos] == '{':
                            lbraces += 1
                        elif after[pos] == '}':
                            rbraces += 1
                        if lbraces == rbraces:
                            break
                        pos += 1
                        if pos == len(after):
                            texlinenum += 1
                            after += tex[texlinenum]
                    mainarg = after[1:pos]
                    after = after[pos+1:]
                else:
                    # Deal with matched character delims
                    delim = after[0]
                    while after.count(delim) < 2:
                        texlinenum += 1
                        after += tex[texlinenum]
                    mainarg, after = after[1:].split(delim, 1)
                arglist.append(mainarg)


        # Do substitution, depending on what is required
        # Need a variable for processed content to be added to texout
        processed = None
        if depy_typeset == 'c':
            if depy_type == 'cmd':
                # Make sure that `after` contains two lines of content
                # This is needed for some replacements that need to look ahead
                if after.count('\n') < 2:
                    texlinenum += 1
                    after += tex[texlinenum]
                processed, texcontent = replace_code_cmd(depy_name, arglist,
                                                         depy_linenum,
                                                         code_replacement,
                                                         code_replacement_mode,
                                                         after, depy_lexer,
                                                         firstnumber)
            else:  # depy_type == 'env'
                end_environment = r'\end{' + depy_name + '}'
                if code_replacement is None:
                    if end_environment not in after:
                        while True:
                            texlinenum += 1
                            after += tex[texlinenum]
                            if end_environment in tex[texlinenum]:
                                break
                    code_replacement, after = after.split(end_environment, 1)
                    # If there's content on the line with the end-environment
                    # command, it should be discarded, to imitate TeX
                    if not code_replacement.endswith('\n'):
                        code_replacement = code_replacement.rsplit('\n', 1)[0] + '\n'
                    # Take care of `gobble`
                    if settings['gobble'] == 'auto':
                        code_replacement = textwrap.dedent(code_replacement)
                else:
                    if end_environment not in after:
                        while True:
                            texlinenum += 1
                            if end_environment in tex[texlinenum]:
                                after = tex[texlinenum]
                                break
                    after = after.split(end_environment, 1)[1]
                # Make sure that `after` contains two lines of content
                # This is needed for some replacements that need to look ahead
                if after.count('\n') < 2:
                    texlinenum += 1
                    after += tex[texlinenum]
                processed, texcontent = replace_code_env(depy_name, arglist,
                                                         depy_linenum,
                                                         code_replacement,
                                                         code_replacement_mode,
                                                         after, depy_lexer,
                                                         firstnumber)
        elif depy_typeset == 'p' and print_replacement is not None:
            if depy_type == 'cmd':
                # Make sure that `after` contains two lines of content
                # This is needed for some replacements that need to look ahead
                if after.count('\n') < 2:
                    texlinenum += 1
                    after += tex[texlinenum]
                processed, texcontent = replace_print_cmd(depy_name, arglist,
                                                          depy_linenum,
                                                          print_replacement,
                                                          print_replacement_mode,
                                                          source,
                                                          after)
            else:  # depy_type == 'env'
                end_environment = r'\end{' + depy_name + '}'
                if end_environment not in after:
                    while True:
                        texlinenum += 1
                        if end_environment in tex[texlinenum]:
                            after = tex[texlinenum]
                            break
                after = after.split(end_environment, 1)[1]
                # Make sure that `after` contains two lines of content
                # This is needed for some replacements that need to look ahead
                if after.count('\n') < 2:
                    texlinenum += 1
                    after += tex[texlinenum]
                processed, texcontent = replace_print_env(depy_name, arglist,
                                                          depy_linenum,
                                                          print_replacement,
                                                          print_replacement_mode,
                                                          source,
                                                          after)
        else:  # depy_typeset == 'n' or (depy_typeset == 'p' and print_replacement is None):
            if depy_type == 'cmd':
                texcontent = after
            else:  # depy_type == 'env'
                end_environment = r'\end{' + depy_name + '}'
                if end_environment not in after:
                    while True:
                        texlinenum += 1
                        if end_environment in tex[texlinenum]:
                            after = tex[texlinenum]
                            break
                after = after.split(end_environment, 1)[1]
                if bool(match('\s*\n', after)):
                    # If the line following `after` is whitespace, it should
                    # be stripped, since most environments throw away
                    # anything after the end of the environment
                    after = after.split('\n')[1]
                texcontent = after
        # #### Once it's supported on the TeX side, need to add support for
        # pc and cp


        # Store any processed content
        if processed is not None:
            texout.append(processed)


# Transfer anything that's left in tex to texout
texout.append(texcontent)
texout.extend(tex[texlinenum+1:])




# Replace the `\usepackage{pythontex}`
for n, line in enumerate(texout):
    if '{pythontex}' in line:
        startline = n
        while '\\usepackage' not in texout[startline] and startline >= 0:
            startline -= 1
        if startline == n:
            if bool(search(r'\\usepackage(?:\[.*?\]){0,1}\{pythontex\}', line)):
                texout[n] = sub(r'\\usepackage(?:\[.*?\]){0,1}\{pythontex\}', '', line)
                if texout[n].isspace():
                    texout[n] = ''
                break
        else:
            content = ''.join(texout[startline:n+1])
            if bool(search(r'(?s)\\usepackage(?:\[.*?\]\s*){0,1}\{pythontex\}', content)):
                replacement = sub(r'(?s)\\usepackage(?:\[.*?\]\s*){0,1}\{pythontex\}', '', content)
                if replacement.isspace():
                    replacement = ''
                texout[startline] = replacement
                for l in range(startline+1, n+1):
                    texout[l] = ''
                break
    elif line.startswith(r'\begin{document}'):
        break
if preamble_additions:
    texout[n] += '\n'.join(preamble_additions) + '\n'
# Take care of graphicspath
if args.graphicspath and settings['graphicx']:
    for n, line in enumerate(texout):
        if '\\graphicspath' in line and not bool(match('\s*%', line)):
            texout[n] = line.replace('\\graphicspath{', '\\graphicspath{{' + settings['outputdir'] +'/}')
            break
        elif line.startswith(r'\begin{document}'):
            texout[n] = '\\graphicspath{{' + settings['outputdir'] + '/}}\n' + line
            break




# Print any final messages
if forced_double_space_list:
    print('* DePythonTeX warning:')
    print('    A trailing double space was forced with "\\space{}" for the following')
    print('    This can happen when printed content is included inline')
    print('    The forced double space is only an issue if it is not intentional')
    for name, linenum in forced_double_space_list:
        print('      "' + name + '" near line ' + str(linenum))




# Write output
if args.output is not None:
    for line in texout:
        outfile.write(line)
    outfile.close()
else:
    if sys.version_info[0] == 2:
        sys.stdout = codecs.getwriter(encoding)(sys.stdout, 'strict')
        sys.stderr = codecs.getwriter(encoding)(sys.stderr, 'strict')
    else:
        sys.stdout = codecs.getwriter(encoding)(sys.stdout.buffer, 'strict')
        sys.stderr = codecs.getwriter(encoding)(sys.stderr.buffer, 'strict')
    for line in texout:
        sys.stdout.write(line)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Install PythonTeX

This installation script is written to work with TeX Live and MiKTeX.  Note
that PythonTeX is included in TeX Live 2013 and later, and may be installed 
via the package manager.  Thus, this installation script is only needed with 
TeX Live when you wish to install the latest version.  PythonTeX is not 
currently available via the MiKTeX package manager.

The script will automatically overwrite (and thus update) all previously 
installed PythonTeX files in the designated installation location.  When 
Kpathsea is available, files may be installed in TEXMFDIST, TEXMFLOCAL, 
TEXMFHOME, or a manually specified location.  Otherwise, the installation 
location must be specified manually.  Installing in TEXMFDIST is useful 
under TeX Live if you want to install PythonTeX and then update it in the 
future via the package manager.

The `mktexlsr` (TeX Live) or `initexmf --update-fndb` (MiKTeX) command is 
executed at the end of the script, to make the system aware of any new files.

Under TeX Live, the script attempts to create a binary wrapper (Windows) or 
symlink (Linux and OS X) for launching the main PythonTeX scripts, 
`pythontex*.py` and `depythontex*.py`.  Under MiKTeX, it attempts to create
a batch file in `miktex/bin`.


Copyright (c) 2012-2014, Geoffrey M. Poore
All rights reserved.
Licensed under the BSD 3-Clause License:
    http://www.opensource.org/licenses/BSD-3-Clause

'''


# Imports
import sys
import platform
from os import path, mkdir, makedirs
if platform.system() != 'Windows':
    # Only create symlinks if not under Windows 
    # (os.symlink doesn't exist under Windows)
    from os import symlink, chmod, unlink
from subprocess import call, check_call, check_output
from shutil import copy
import textwrap


# We need a version of input that works under both Python 2 and 3
try:
    input = raw_input
except:
    pass


# Print startup messages and notices
print('Preparing to install PythonTeX')
if platform.system() != 'Windows':
    message = '''
              You may need to run this script with elevated permissions
              and/or specify the environment.  For example, you may need
              "sudo env PATH=$PATH".  That is typically necessary when your
              system includes a TeX distribution, and you have manually
              installed another distribution (common with Ubuntu etc.).  If 
              the installation path you want is not automatically detected, 
              it may indicate a permissions issue.              
              '''
    print(textwrap.dedent(message))


# Attempt to detect the TeX distribution
try:
    if sys.version_info.major == 2:
        texout = check_output(['latex', '--version'])
    else:
        texout = check_output(['latex', '--version']).decode('utf-8')
except:
    sys.exit('Could not retrieve latex info when running "latex --version"')
if 'TeX Live' in texout:
    detected_texdist = True
    texlive = True
    miktex = False
elif platform.system() == 'Windows' and 'MiKTeX' in texout:
    detected_texdist = True
    texlive = False
    miktex = True
else:
    detected_texdist = False
    texlive = False
    miktex = False


# Make sure all necessary files are present
# The pythontex_gallery and pythontex_quickstart are optional; we 
# check for them when installing doc, and install if available
needed_files = ['pythontex.py', 'pythontex2.py', 'pythontex3.py',
                'pythontex_engines.py', 'pythontex_utils.py',
                'depythontex.py', 'depythontex2.py', 'depythontex3.py',
                'pythontex.sty', 'pythontex.ins', 'pythontex.dtx', 
                'pythontex.pdf', 'README',
                'syncpdb.py']
missing_files = False
# Print a list of all files that are missing, and exit if any are
for eachfile in needed_files:
    if not path.exists(eachfile):
        print('Could not find file ' + eachfile)
        missing_files = True
if missing_files:
    sys.exit('Exiting due to missing files.')


# Retrieve the location of valid TeX trees
if sys.version_info[0] == 2:
    try:
        texmf_dist = check_output(['kpsewhich', '-var-value', 'TEXMFDIST']).rstrip('\r\n')
    except:
        texmf_dist = None
    try:
        texmf_local = check_output(['kpsewhich', '-var-value', 'TEXMFLOCAL']).rstrip('\r\n')
    except:
        texmf_local = None
    try:
        texmf_home = check_output(['kpsewhich', '-var-value', 'TEXMFHOME']).rstrip('\r\n')
    except:
        texmf_home = None
else:
    try:
        texmf_dist = check_output(['kpsewhich', '-var-value', 'TEXMFDIST']).decode('utf-8').rstrip('\r\n')
    except:
        texmf_dist = None
    try:
        texmf_local = check_output(['kpsewhich', '-var-value', 'TEXMFLOCAL']).decode('utf-8').rstrip('\r\n')
    except:
        texmf_local = None
    try:
        texmf_home = check_output(['kpsewhich', '-var-value', 'TEXMFHOME']).decode('utf-8').rstrip('\r\n')
    except:
        texmf_home = None


# Get installation location from user
texmf_vars = [texmf_dist, texmf_local, texmf_home]
message = '''
          Choose an installation location.
          
          TEXMFDIST is a good choice if you want to update PythonTeX 
          in the future using your TeX distribution's package manager
          (assuming that is supported).
          
            1. TEXMFDIST
                 {0}
            2. TEXMFLOCAL
                 {1}
            3. TEXMFHOME
                 {2}
            4. Manual location
            
            5. Exit without installing
          '''.format(*[x if x else '<INVALID>' for x in texmf_vars])

if any(texmf_vars):
    path_choice = ''
    while (path_choice not in ('1', '2', '3', '4', '5') or 
            (int(path_choice) <= 3 and not texmf_vars[int(path_choice)-1])):
        print(textwrap.dedent(message))
        path_choice = input('Installation location (number):  ')
        if path_choice == '':
            sys.exit()
    if path_choice == '1':
        texmf_path = texmf_dist
    elif path_choice == '2':
        texmf_path = texmf_local
    elif path_choice == '3':
        texmf_path = texmf_home
    elif path_choice == '4':
        texmf_path = input('Enter a path:\n')
        if texmf_path == '':
            sys.exit()
        if platform.system() == 'Windows':
            if 'texlive' in texmf_path.lower():
                detected_texdist = True
                texlive = True
                miktex = False
            elif 'miktex' in texmf_path.lower():
                detected_texdist = True
                texlive = False
                miktex = True
    else:
        sys.exit()
else:
    print('Failed to detect possible installation locations automatically.')
    print('TEXMF paths could not be located with kpsewhich.')
    texmf_path = input('Plese enter an installation path, or press "Enter" to exit:\n')
    if texmf_path == '':
        sys.exit()

# Make sure path slashes are compatible with the operating system
# Kpathsea returns forward slashes, but Windows needs back slashes
texmf_path = path.expandvars(path.expanduser(path.normcase(texmf_path)))

# Check to make sure the path is valid 
# This should only be needed for manual input, but it's a good check
if not path.isdir(texmf_path):
    sys.exit('Invalid installation path.  Exiting.')

# Now check that all other needed paths are present
if path_choice != '2':
    doc_path = path.join(texmf_path, 'doc', 'latex')
    package_path = path.join(texmf_path, 'tex', 'latex')
    scripts_path = path.join(texmf_path, 'scripts')
    source_path = path.join(texmf_path, 'source', 'latex')
else:
    doc_path = path.join(texmf_path, 'doc', 'latex', 'local')
    package_path = path.join(texmf_path, 'tex', 'latex', 'local')
    scripts_path = path.join(texmf_path, 'scripts', 'local')
    source_path = path.join(texmf_path, 'source', 'latex', 'local')
# May need to create some local directories
make_paths = False
for eachpath in [doc_path, package_path, scripts_path, source_path]:
    if not path.exists(eachpath):
        if make_paths:
            makedirs(eachpath)
            print('  * Created ' + eachpath)
        else:
            choice = ''
            while choice not in ('y', 'n'):
                choice = input('Some directories do not exist.  Create them? [y/n]  ')
                if choice == '':
                    sys.exit()
            if choice == 'y':
                make_paths = True
                try:
                    makedirs(eachpath)
                    print('  * Created ' + eachpath)
                except (OSError, IOError) as e:
                    if e.errno == 13:
                        print('\nInsufficient permission to install PythonTeX')
                        if platform.system() == 'Windows':
                            message = '''
                                      You may need to run the installer as "administrator".
                                      This may be done under Vista and later by right-clicking on
                                      pythontex_install.bat, then selecting "Run as administrator".
                                      Or you can open a command prompt as administrator 
                                      (Start, Programs, Accessories, right-click Command Prompt,
                                      Run as administrator), change to the directory in which
                                      pythontex_install.py is located, and run 
                                      "python pythontex_install.py".
                                      '''
                            print(textwrap.dedent(message))
                            call(['pause'], shell=True)
                        else:
                            print('(For example, you may need "sudo", or possibly "sudo env PATH=$PATH")\n')
                        sys.exit(1)
                    else:
                        raise
            else:
                message = '''
                          Paths were not created.  The following will be needed.
                            * {0}
                            * {1}
                            * {2}
                            * {3}
                          
                          Exiting.
                          '''.format(doc_path, package_path, scripts_path, source_path)
                print(textwrap.dedent(message))
                sys.exit()

# Modify the paths by adding the pythontex directory, which will be created
doc_path = path.join(doc_path, 'pythontex')
package_path = path.join(package_path, 'pythontex')
scripts_path = path.join(scripts_path, 'pythontex')
source_path = path.join(source_path, 'pythontex')


# Install files
# Use a try/except in case elevated permissions are needed (Linux and OS X)
print('\nPythonTeX will be installed in \n  ' + texmf_path)
try:
    # Install docs
    if not path.exists(doc_path):
        mkdir(doc_path)
    copy('pythontex.pdf', doc_path)
    copy('README', doc_path)
    for doc in ('pythontex_quickstart.tex', 'pythontex_quickstart.pdf', 
                'pythontex_gallery.tex', 'pythontex_gallery.pdf'):
        if path.isfile(doc):
            copy(doc, doc_path)
        else:
            doc = path.join('..', doc.rsplit('.', 1)[0], doc)
            if path.isfile(doc):
                copy(doc, doc_path)
    # Install package
    if not path.exists(package_path):
        mkdir(package_path)
    copy('pythontex.sty', package_path)
    # Install scripts
    if not path.exists(scripts_path):
        mkdir(scripts_path)
    copy('pythontex.py', scripts_path)
    copy('depythontex.py', scripts_path)
    copy('pythontex_utils.py', scripts_path)
    copy('pythontex_engines.py', scripts_path)
    copy('syncpdb.py', scripts_path)
    for ver in [2, 3]:
        copy('pythontex{0}.py'.format(ver), scripts_path)
        copy('depythontex{0}.py'.format(ver), scripts_path)
    # Install source
    if not path.exists(source_path):
        mkdir(source_path)
    copy('pythontex.ins', source_path)
    copy('pythontex.dtx', source_path)
except (OSError, IOError) as e:
    if e.errno == 13:
        print('\nInsufficient permission to install PythonTeX')
        if platform.system() == 'Windows':
            message = '''
                      You may need to run the installer as "administrator".
                      This may be done under Vista and later by right-clicking on
                      pythontex_install.bat, then selecting "Run as administrator".
                      Or you can open a command prompt as administrator 
                      (Start, Programs, Accessories, right-click Command Prompt,
                      Run as administrator), change to the directory in which
                      pythontex_install.py is located, and run 
                      "python pythontex_install.py".
                      '''
            print(textwrap.dedent(message))
            call(['pause'], shell=True)
        else:
            print('(For example, you may need "sudo", or possibly "sudo env PATH=$PATH")\n')
        sys.exit(1)
    else:
        raise        


# Install binary wrappers, create symlinks, or suggest the creation of 
# wrappers/batch files/symlinks.  This part is operating system dependent.
if platform.system() == 'Windows':
    # If under Windows, we create a binary wrapper if under TeX Live 
    # or a batch file if under MiKTeX.  Otherwise, alert the user 
    # regarding the need for a wrapper or batch file.    
    if miktex:
        try:
            if sys.version_info.major == 2:
                bin_path = check_output(['kpsewhich', '-var-value', 'TEXMFDIST']).rstrip('\r\n')
            else:
                bin_path = check_output(['kpsewhich', '-var-value', 'TEXMFDIST']).decode('utf-8').rstrip('\r\n')
            bin_path = path.join(bin_path, 'miktex', 'bin')
            
            for s in ('pythontex.py', 'depythontex.py'):
                batch = '@echo off\n"{0}" %*\n'.format(path.join(scripts_path, s))
                f = open(path.join(bin_path, s.replace('.py', '.bat')), 'w')
                f.write(batch)
                f.close()
        except:
            message = '''
                      Could not create a batch file for launching pythontex.py and 
                      depythontex.py.  You will need to create a batch file manually.
                      Sample batch files are included with the main PythonTeX files.
                      The batch files should be in a location on the Windows PATH.
                      The bin/ directory in your TeX distribution may be a good 
                      location.
                      
                      The scripts pythontex.py and depythontex.py are located in 
                      the following directory:
                        {0}
                      '''.format(scripts_path)
            print(textwrap.dedent(message))
    else:
        # Assemble the binary path, assuming TeX Live
        # The directory bin/ should be at the same level as texmf
        bin_path = path.join(path.split(texmf_path)[0], 'bin', 'win32') 
        if path.exists(path.join(bin_path, 'runscript.exe')):
            for f in ('pythontex.py', 'depythontex.py'):
                copy(path.join(bin_path, 'runscript.exe'), path.join(bin_path, '{0}.exe'.format(f.rsplit('.')[0])))
            print('\nCreated binary wrapper...')
        else:
            message = '''
                      Could not create a wrapper for launching pythontex.py and 
                      depythontex.py; did not find runscript.exe.  You will need 
                      to create a wrapper manually, or use a batch file.  Sample 
                      batch files are included with the main PythonTeX files.  
                      The wrapper or batch file should be in a location on the 
                      Windows PATH.  The bin/ directory in your TeX distribution 
                      may be a good location.
                      
                      The scripts pythontex.py and depythontex.py are located in 
                      the following directory:
                        {0}
                      '''.format(scripts_path)
            print(textwrap.dedent(message))
else:
    # Optimistically proceed as if every system other than Windows can 
    # share one set of code.
    root_path = path.split(texmf_path)[0]
    # Create a list of all possible subdirectories of bin/ for TeX Live
    # Source:  http://www.tug.org/texlive/doc/texlive-en/texlive-en.html#x1-250003.2.1
    texlive_platforms = ['alpha-linux', 'amd64-freebsd', 'amd64-kfreebsd',
                         'armel-linux', 'i386-cygwin', 'i386-freebsd',
                         'i386-kfreebsd', 'i386-linux', 'i386-solaris',
                         'mips-irix', 'mipsel-linux', 'powerpc-aix', 
                         'powerpc-linux', 'sparc-solaris', 'universal-darwin',
                         'x86_64-darwin', 'x86_64-linux', 'x86_64-solaris']
    symlink_created = False
    # Try to create a symlink in the standard TeX Live locations
    for pltfrm in texlive_platforms:
        bin_path = path.join(root_path, 'bin', pltfrm)
        if path.exists(bin_path):
            # Unlink any old symlinks if they exist, and create new ones
            # Not doing this gave permissions errors under Ubuntu
            for f in ('pythontex.py', 'pythontex2.py', 'pythontex3.py',
                      'depythontex.py', 'depythontex2.py', 'depythontex3.py'):
                link = path.join(bin_path, f)
                if path.exists(link):
                    unlink(link)
                symlink(path.join(scripts_path, f), link)
                chmod(link, 0o775)
            symlink_created = True
    
    # If the standard TeX Live bin/ locations didn't work, try the typical 
    # location for MacPorts TeX Live.  This should typically be 
    # /opt/local/bin, but instead of assuming that location, we just climb 
    # two levels up from texmf-dist and then look for a bin/ directory that
    # contains a tex executable.  (For MacPorts, texmf-dist should be at 
    # /opt/local/share/texmf-dist.)
    if not symlink_created and platform.system() == 'Darwin':
        bin_path = path.join(path.split(root_path)[0], 'bin')
        if path.exists(bin_path):
            try:
                # Make sure this bin/ is the bin/ we're looking for, by
                # seeing if pdftex exists
                check_output([path.join(bin_path, 'pdftex'), '--version'])
                # Create symlinks
                for f in ('pythontex.py', 'pythontex2.py', 'pythontex3.py',
                          'depythontex.py', 'depythontex2.py', 'depythontex3.py'):
                    link = path.join(bin_path, f)
                    if path.exists(link):
                        unlink(link)
                    symlink(path.join(scripts_path, f), link)
                    chmod(link, 0o775)
                symlink_created = True
            except:
                pass
    if symlink_created:
        print("\nCreated symlink in Tex's bin/ directory...")
    else:
        print('\nCould not automatically create a symlink to pythontex*.py and depythontex*.py.')
        print('You may wish to create one manually, and make it executable via chmod.')
        print('The scripts pythontex*.py and depythontex*.py are located in the following directory:')
        print('    ' + scripts_path)


# Alert TeX to the existence of the package via mktexlsr
if not miktex:
    try:
        # Need to adjust if under Windows with a user-specified TeX Live
        # installation and a default MiKTeX installation; want to call
        # mktexlsr for the user-specified TeX Live installation
        if platform.system() == 'Windows' and 'MiKTeX' in texout:
            check_call(path.join(bin_path, 'mktexlsr'))
        else:
            check_call(['mktexlsr'])
        print('\nRunning "mktexlsr" to make TeX aware of new files...')
    except:
        print('Could not run "mktexlsr".')
        print('Your system may not be aware of newly installed files.')
else:
    success = False
    try:
        check_call(['initexmf', '--admin', '--update-fndb'])
        print('\nRunning "initexmf --admin --update-fndb" to make TeX aware of new files...')
        check_call(['initexmf', '--update-fndb'])
        print('\nRunning "initexmf --update-fndb" to make TeX aware of new files...')
        success = True
    except:
        pass
    if not success:
        try:
            check_call(['initexmf', '--update-fndb'])
            print('\nRunning "initexmf --update-fndb" to make TeX aware of new files...')
            print('Depending on your installation settings, you may also need to run')
            print('"initexmf --admin --update-fndb"')
        except:
            print('Could not run "initexmf --update-fndb" or "initexmf --admin --update-fndb"')
            print('Your system may not be aware of newly installed files.')

            
if platform.system() == 'Windows':
    # Pause so that the user can see any errors or other messages
    # input('\n[Press ENTER to exit]')
    print('\n')
    call(['pause'], shell=True)



#!/usr/bin/env python3
###########################################################################
#
# xasyOptions provides a mechanism for storing and restoring a user's
# preferences.
#
#
# Author: Orest Shardt
# Created: June 29, 2007
#
###########################################################################

import sys
import io
import os
import platform
import shutil
import configs
import cson

class xasyOptions:
    def defaultOptions(self):
        if self._defaultOptions is None:
            f = io.open(self._defaultOptLocation)
            try:
                opt = cson.loads(f.read())
            finally:
                f.close()
            self._defaultOptions = opt
        return self._defaultOptions

    def overrideSettings(self):
        settingsName = platform.system()

        if settingsName not in self.options:
            return

        for key in self.options[settingsName]:
            self.options[key] = self.options[settingsName][key]


    def settingsFileLocation(self):
        folder = os.path.expanduser("~/.asy/")

        searchOrder = ['.cson', '']

        searchIndex = 0
        found = False
        currentFile = ''
        while searchIndex < len(searchOrder) and not found:
            currentFile = os.path.join(folder, self.configName + searchOrder[searchIndex])
            if os.path.isfile(currentFile):
                found = True
            searchIndex += 1

        if found:
            return os.path.normcase(currentFile)
        else:
            return os.path.normcase(os.path.join(folder, self.configName + '.cson'))

    def __init__(self, configName, defaultConfigLocation):
        self.configName = configName
        self.defaultConfigName = defaultConfigLocation

        self._defaultOptions = None
        self._defaultOptLocation = os.path.join(defaultConfigLocation)

        self.options = self.defaultOptions()
        self.load()

    def __getitem__(self, key):
        return self.options[key]

    def __contains__(self, key):
        return key in self.options

    def get(self, key, default=None):
        if key not in self.options:
            return default
        return self.options[key]

    def __setitem__(self, key, value):
        self.options[key] = value

    def load(self):
        fileName = self.settingsFileLocation()
        if not os.path.exists(fileName):
            # make folder
            thedir = os.path.dirname(fileName)
            if not os.path.exists(thedir):
                os.makedirs(thedir)
            if not os.path.isdir(thedir):
                raise Exception("Configuration folder path does not point to a folder")
            self.setDefaults()
        f = io.open(fileName, 'r')
        try:
            ext = os.path.splitext(fileName)[1]
            newOptions = cson.loads(f.read())
        except (IOError, ModuleNotFoundError):
            self.setDefaults()
        else:
            for key, val in self.options.items():
                if key in newOptions:
                    if val is not None:
                        assert isinstance(newOptions[key], type(val))
                else:
                    newOptions[key] = self.options[key]
            self.options = newOptions
        finally:
            f.close()
        self.overrideSettings()

    def setDefaults(self):
        self.options = self.defaultOptions()
        if sys.platform[:3] == 'win':  # for windows, wince, win32, etc
            # setAsyPathFromWindowsRegistry()
            pass
        folder = os.path.expanduser("~/.asy/")
        defaultPath = os.path.join(folder, self.configName + '.cson')
        shutil.copy2(self._defaultOptLocation, defaultPath)


# TODO: Figure out how to merge this back.
"""
def setAsyPathFromWindowsRegistry():
    if os.name == 'nt':
        import _winreg as registry
        # test both registry locations
        try:
            key = registry.OpenKey(registry.HKEY_LOCAL_MACHINE,
                                   "Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\Asymptote")
            options['asyPath'] = registry.QueryValueEx(key, "Path")[0] + "\\asy.exe"
            registry.CloseKey(key)
        except:
            key = registry.OpenKey(registry.HKEY_LOCAL_MACHINE,
                                   "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Asymptote")
            options['asyPath'] = registry.QueryValueEx(key, "InstallLocation")[0] + "\\asy.exe"
            registry.CloseKey(key)
"""

class xasyOpenRecent:
    def __init__(self, configName, defaultConfigLocation):
        self.configName = configName
        self.fileName = self.settingsFileLocation()
        if not os.path.isfile(self.fileName):
            f = io.open(self.fileName, 'w')
            f.write('')
            f.close()

    def settingsFileLocation(self):
        folder = os.path.expanduser("~/.asy/")

        currentFile = os.path.join(folder, self.configName + '.txt')
        return os.path.normcase(currentFile)

    def insert(self, path):
        if not os.path.exists(self.fileName):
            # make folder
            thedir = os.path.dirname(self.fileName)
            if not os.path.exists(thedir):
                os.makedirs(thedir)
            if not os.path.isdir(thedir):
                raise Exception("Configuration folder path does not point to a folder")

        f = io.open(self.fileName, 'r')
        lines = f.readlines()
        f.close()

        f = io.open(self.fileName, 'w')
        f.write(path.strip() + '\n')
        for line in lines:
            if line.strip() != path.strip():
                f.write(line.strip() + '\n')
        f.close()

    @property
    def pathList(self):
        self.findingPaths=True
        return self.findPath()

    def findPath(self):
        f = io.open(self.fileName, 'r')
        paths = [path.strip() for path in f.readlines()]
        f.close()

        trueFiles = list(map(lambda path: os.path.isfile(os.path.expanduser(path)), paths))
        if all(trueFiles):
            return paths
        else:
            if not self.findingPaths:
                raise RecursionError
            self.findingPaths = False
            self.removeNotFound(list(trueFiles), paths)
            return self.findPath()

    def removeNotFound(self, trueFiles, paths):
        f = io.open(self.fileName, 'w')
        for index, path in enumerate(paths):
            if trueFiles[index]:
                f.write(path + '\n')
        f.close()

    def clear(self):
        f = io.open(self.fileName, 'w')
        f.write('')
        f.close()

class BasicConfigs:
    _configPath = list(configs.__path__)[0]
    defaultOpt = xasyOptions(
        'xasyconfig', os.path.join(_configPath, 'xasyconfig.cson'))
    keymaps = xasyOptions('xasykeymap', os.path.join(
        _configPath, 'xasykeymap.cson'))
    openRecent = xasyOpenRecent('xasyrecents', os.path.join( _configPath, "xasyrecent.txt"))
